import hashlib
import plistlib
from pathlib import Path
from typing import List, Mapping

from pymobiledevice3.exceptions import AlreadyMountedError, DeveloperModeIsNotEnabledError, InternalError, \
    MessageNotSupportedError, MissingManifestError, NotMountedError, PyMobileDevice3Exception, \
    UnsupportedCommandError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.restore.tss import TSSRequest
from pymobiledevice3.services.base_service import BaseService


class MobileImageMounterService(BaseService):
    # implemented in /usr/libexec/mobile_storage_proxy
    SERVICE_NAME = 'com.apple.mobile.mobile_image_mounter'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def copy_devices(self) -> List[Mapping]:
        """ Copy mounted devices list. """
        try:
            return self.service.send_recv_plist({'Command': 'CopyDevices'})['EntryList']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def lookup_image(self, image_type: str) -> bytes:
        """ Lookup mounted image by its name. """
        response = self.service.send_recv_plist({'Command': 'LookupImage',
                                                 'ImageType': image_type})

        if not response or not response.get('ImagePresent', True):
            raise NotMountedError()

        signature = response.get('ImageSignature', [])
        if isinstance(signature, list):
            if not signature:
                raise NotMountedError()
            return signature[0]
        return signature

    def is_image_mounted(self, image_type: str) -> bool:
        try:
            self.lookup_image(image_type)
            return True
        except NotMountedError:
            return False

    def unmount_image(self, mount_path: str) -> None:
        """ umount image (Added on iOS 14.0) """
        request = {'Command': 'UnmountImage', 'MountPath': mount_path}
        response = self.service.send_recv_plist(request)

        error = response.get('Error')
        if error:
            if error == 'UnknownCommand':
                raise UnsupportedCommandError()
            elif 'There is no matching entry' in response.get('DetailedError', ''):
                raise NotMountedError(response)
            elif error == 'InternalError':
                raise InternalError(response)
            else:
                raise PyMobileDevice3Exception(response)

    def mount_image(self, image_type: str, signature: bytes, extras: Mapping = None) -> None:
        """ Upload image into device. """

        if self.is_image_mounted(image_type):
            raise AlreadyMountedError()

        request = {'Command': 'MountImage',
                   'ImageType': image_type,
                   'ImageSignature': signature}

        if extras is not None:
            request.update(extras)
        response = self.service.send_recv_plist(request)

        if 'Developer mode is not enabled' in response.get('DetailedError', ''):
            raise DeveloperModeIsNotEnabledError()

        status = response.get('Status')

        if status != 'Complete':
            raise PyMobileDevice3Exception(f'command MountImage failed with: {response}')

    def upload_image(self, image_type: str, image: bytes, signature: bytes) -> None:
        """ Upload image into device. """
        self.service.send_plist({'Command': 'ReceiveBytes',
                                 'ImageType': image_type,
                                 'ImageSize': len(image),
                                 'ImageSignature': signature})
        result = self.service.recv_plist()

        status = result.get('Status')

        if status != 'ReceiveBytesAck':
            raise PyMobileDevice3Exception(f'command ReceiveBytes failed with: {result}')

        self.service.sendall(image)
        result = self.service.recv_plist()

        status = result.get('Status')

        if status != 'Complete':
            raise PyMobileDevice3Exception(f'command ReceiveBytes failed to send bytes with: {result}')

    def query_developer_mode_status(self) -> bool:
        response = self.service.send_recv_plist({'Command': 'QueryDeveloperModeStatus'})

        try:
            return response['DeveloperModeStatus']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def query_nonce(self, personalized_image_type: str = None) -> bytes:
        request = {'Command': 'QueryNonce'}
        if personalized_image_type is not None:
            request['PersonalizedImageType'] = personalized_image_type
        response = self.service.send_recv_plist(request)
        try:
            return response['PersonalizationNonce']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def query_personalization_identifiers(self, image_type: str = None) -> Mapping:
        request = {'Command': 'QueryPersonalizationIdentifiers'}

        if image_type is not None:
            request['PersonalizedImageType'] = image_type

        response = self.service.send_recv_plist(request)

        try:
            return response['PersonalizationIdentifiers']
        except KeyError as e:
            raise MessageNotSupportedError from e

    def query_personalization_manifest(self, image_type: str, signature: bytes) -> bytes:
        response = self.service.send_recv_plist({
            'Command': 'QueryPersonalizationManifest', 'PersonalizedImageType': image_type, 'ImageType': image_type,
            'ImageSignature': signature})
        try:
            # The response "ImageSignature" is actually an IM4M
            return response['ImageSignature']
        except KeyError:
            raise MissingManifestError()

    def roll_personalization_nonce(self) -> None:
        try:
            self.service.send_recv_plist({'Command': 'RollPersonalizationNonce'})
        except ConnectionAbortedError:
            return

    def roll_cryptex_nonce(self) -> None:
        try:
            self.service.send_recv_plist({'Command': 'RollCryptexNonce'})
        except ConnectionAbortedError:
            return


class DeveloperDiskImageMounter(MobileImageMounterService):
    IMAGE_TYPE = 'Developer'

    def mount(self, image: Path, signature: Path) -> None:
        if self.is_image_mounted(self.IMAGE_TYPE):
            raise AlreadyMountedError()

        image = Path(image).read_bytes()
        signature = Path(signature).read_bytes()
        self.upload_image(self.IMAGE_TYPE, image, signature)
        self.mount_image(self.IMAGE_TYPE, signature)

    def umount(self) -> None:
        self.unmount_image('/Developer')


class PersonalizedImageMounter(MobileImageMounterService):
    IMAGE_TYPE = 'Personalized'

    def mount(self, image: Path, build_manifest: Path, trust_cache: Path,
              info_plist: Mapping = None) -> None:
        if self.is_image_mounted(self.IMAGE_TYPE):
            raise AlreadyMountedError()

        image = image.read_bytes()
        trust_cache = trust_cache.read_bytes()

        # try to fetch the personalization manifest if the device already has one
        # in case of failure, the service will close the socket, so we'll have to reestablish the connection
        # and query the manifest from Apple's ticket server instead
        try:
            manifest = self.query_personalization_manifest('DeveloperDiskImage', hashlib.sha384(image).digest())
        except MissingManifestError:
            self.service = self.lockdown.start_service(self.SERVICE_NAME)
            manifest = self.get_manifest_from_tss(plistlib.loads(build_manifest.read_bytes()))

        self.upload_image(self.IMAGE_TYPE, image, manifest)

        extras = {}
        if info_plist is not None:
            extras['ImageInfoPlist'] = info_plist
        extras['ImageTrustCache'] = trust_cache
        self.mount_image(self.IMAGE_TYPE, manifest, extras=extras)

    def umount(self) -> None:
        self.unmount_image('/System/Developer')

    def get_manifest_from_tss(self, build_manifest: Mapping) -> bytes:
        request = TSSRequest()

        personalization_identifiers = self.query_personalization_identifiers()
        for key, value in personalization_identifiers.items():
            if key.startswith('Ap,'):
                request.update({key: value})

        board_id = personalization_identifiers['BoardId']
        chip_id = personalization_identifiers['ChipID']

        build_identity = None
        for tmp_build_identity in build_manifest['BuildIdentities']:
            if int(tmp_build_identity['ApBoardID'], 0) == board_id and \
                    int(tmp_build_identity['ApChipID'], 0) == chip_id:
                build_identity = tmp_build_identity
                break
        manifest = build_identity['Manifest']

        parameters = {
            'ApProductionMode': True,
            'ApSecurityDomain': 1,
            'ApSecurityMode': True,
            'ApSupportsImg4': True,
        }

        request.update({
            '@ApImg4Ticket': True,
            '@BBTicket': True,
            'ApBoardID': board_id,
            'ApChipID': chip_id,
            'ApECID': self.lockdown.ecid,
            'ApNonce': self.query_nonce('DeveloperDiskImage'),
            'ApProductionMode': True,
            'ApSecurityDomain': 1,
            'ApSecurityMode': True,
            'SepNonce': b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
            'UID_MODE': False,
        })

        for key, manifest_entry in manifest.items():
            info_dict = manifest_entry.get('Info')
            if info_dict is None:
                continue

            if not manifest_entry.get('Trusted', False):
                self.logger.debug(f'skipping {key} as it is not trusted')
                continue

            # copy this entry
            tss_entry = dict(manifest_entry)

            # remove obsolete Info node
            tss_entry.pop('Info')

            # handle RestoreRequestRules
            if 'RestoreRequestRules' in manifest['LoadableTrustCache']['Info']:
                rules = manifest['LoadableTrustCache']['Info']['RestoreRequestRules']
                if rules:
                    self.logger.debug(f'Applying restore request rules for entry {key}')
                    tss_entry = request.apply_restore_request_rules(tss_entry, parameters, rules)

            # Make sure we have a Digest key for Trusted items even if empty
            if manifest_entry.get('Digest') is None:
                tss_entry['Digest'] = bytes()

            request.update({key: tss_entry})

        response = request.send_receive()
        return response['ApImg4Ticket']



import logging
import select
import socket
import threading

from pymobiledevice3 import usbmux
from pymobiledevice3.exceptions import ConnectionFailedError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.service_connection import ServiceConnection


class TcpForwarder:
    """
    Allows forwarding local tcp connection into the device via a given lockdown connection
    """

    MAX_FORWARDED_CONNECTIONS = 200
    TIMEOUT = 1

    def __init__(self, src_port: int, dst_port: int, serial: str = None, enable_ssl=False,
                 listening_event: threading.Event = None, usbmux_connection_type: str = None):
        """
        Initialize a new tcp forwarder

        :param src_port: tcp port to listen on
        :param dst_port: tcp port to connect to each new connection via the supplied lockdown object
        :param serial: device serial
        :param enable_ssl: enable ssl wrapping for the transferred data
        :param listening_event: event to fire when the listening occurred
        :param usbmux_connection_type: preferred connection type
        """
        self.logger = logging.getLogger(__name__)
        self.serial = serial
        self.src_port = src_port
        self.dst_port = dst_port
        self.server_socket = None
        self.inputs = []
        self.enable_ssl = enable_ssl
        self.stopped = threading.Event()
        self.listening_event = listening_event
        self.usbmux_connection_type = usbmux_connection_type

        # dictionaries containing the required maps to transfer data between each local
        # socket to its remote socket and vice versa
        self.connections = {}

    def start(self, address='0.0.0.0'):
        """ forward each connection from given local machine port to remote device port """
        # create local tcp server socket
        self.server_socket = socket.socket()
        self.server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server_socket.bind((address, self.src_port))
        self.server_socket.listen(self.MAX_FORWARDED_CONNECTIONS)
        self.server_socket.setblocking(False)

        self.inputs = [self.server_socket]
        if self.listening_event:
            self.listening_event.set()

        while self.inputs:
            # will only perform the socket select on the inputs. the outputs will handled
            # as synchronous blocking
            readable, writable, exceptional = select.select(self.inputs, [], self.inputs, self.TIMEOUT)
            if self.stopped.is_set():
                break

            closed_sockets = set()
            for current_sock in readable:
                if current_sock is self.server_socket:
                    self._handle_server_connection()
                else:
                    if current_sock not in closed_sockets:
                        try:
                            self._handle_data(current_sock, closed_sockets)
                        except ConnectionResetError:
                            self._handle_close_or_error(current_sock)

            for current_sock in exceptional:
                self._handle_close_or_error(current_sock)

        # on stop, close all currently opened sockets
        for current_sock in self.inputs:
            current_sock.close()

    def _handle_close_or_error(self, from_sock):
        """ if an error occurred its time to close the two sockets """
        other_sock = self.connections[from_sock]

        other_sock.close()
        from_sock.close()
        self.inputs.remove(other_sock)
        self.inputs.remove(from_sock)

        self.logger.info(f'connection {other_sock} was closed')

    def _handle_data(self, from_sock, closed_sockets):
        data = from_sock.recv(1024)

        if len(data) == 0:
            # no data means socket was closed
            self._handle_close_or_error(from_sock)
            closed_sockets.add(from_sock)
            closed_sockets.add(self.connections[from_sock])
            return

        # when data is received from one end, just forward it to the other
        other_sock = self.connections[from_sock]

        # send the data in blocking manner
        other_sock.setblocking(True)
        other_sock.sendall(data)
        other_sock.setblocking(False)

    def _handle_server_connection(self):
        """ accept the connection from local machine and attempt to connect at remote """
        local_connection, client_address = self.server_socket.accept()
        local_connection.setblocking(False)

        try:
            if self.enable_ssl:
                # use the lockdown pairing record
                lockdown = LockdownClient(self.serial, usbmux_connection_type=self.usbmux_connection_type)
                service_connection = ServiceConnection.create_using_usbmux(
                    self.serial, self.dst_port, connection_type=self.usbmux_connection_type)

                with lockdown.ssl_file() as ssl_file:
                    service_connection.ssl_start(ssl_file)

                remote_connection = service_connection.socket
            else:
                # connect directly using usbmuxd
                mux_device = usbmux.select_device(self.serial, connection_type=self.usbmux_connection_type)
                if mux_device is None:
                    raise ConnectionFailedError()
                remote_connection = mux_device.connect(self.dst_port)
        except ConnectionFailedError:
            self.logger.error(f'failed to connect to port: {self.dst_port}')
            local_connection.close()
            return

        remote_connection.setblocking(False)

        # append the newly created sockets into input list
        self.inputs.append(local_connection)
        self.inputs.append(remote_connection)

        # and store a map of which local connection is transferred to which remote one
        self.connections[remote_connection] = local_connection
        self.connections[local_connection] = remote_connection

        self.logger.info(f'connection established from local to remote port {self.dst_port}')

    def stop(self):
        """ stop forwarding """
        self.stopped.set()



import socket
from typing import Generator, Mapping, Union

from pymobiledevice3.exceptions import NotificationTimeoutError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService


class NotificationProxyService(BaseService):
    SERVICE_NAME = 'com.apple.mobile.notification_proxy'
    INSECURE_SERVICE_NAME = 'com.apple.mobile.insecure_notification_proxy'

    def __init__(self, lockdown: LockdownClient, insecure=False, timeout: Union[float, int] = None):
        if insecure:
            super().__init__(lockdown, self.INSECURE_SERVICE_NAME)
        else:
            super().__init__(lockdown, self.SERVICE_NAME)

        if timeout is not None:
            self.service.socket.settimeout(timeout)

    def notify_post(self, name: str) -> None:
        """ Send notification to the device's notification_proxy. """
        self.service.send_plist({'Command': 'PostNotification',
                                 'Name': name})

    def notify_register_dispatch(self, name: str) -> None:
        """ Tells the device to send a notification on the specified event. """
        self.logger.info(f'Observing {name}')
        self.service.send_plist({'Command': 'ObserveNotification',
                                 'Name': name})

    def receive_notification(self) -> Generator[Mapping, None, None]:
        while True:
            try:
                yield self.service.recv_plist()
            except socket.timeout as e:
                raise NotificationTimeoutError from e



import json
from dataclasses import dataclass
from enum import Enum
from pathlib import Path

RESOURCES = Path(__file__).parent.parent.parent / 'resources' / 'webinspector'
FIND_NODES = (RESOURCES / 'find_nodes.js').read_text()


class By(Enum):
    """
    Set of supported locator strategies.
    """
    ID = 'id'
    XPATH = 'xpath'
    LINK_TEXT = 'link text'
    PARTIAL_LINK_TEXT = 'partial link text'
    NAME = 'name'
    TAG_NAME = 'tag name'
    CLASS_NAME = 'class name'
    CSS_SELECTOR = 'css selector'


class MouseButton(Enum):
    NONE = 'None'
    LEFT = 'Left'
    MIDDLE = 'Middle'
    RIGHT = 'Right'


class MouseInteraction(Enum):
    MOVE = 'Move'
    DOWN = 'Down'
    UP = 'Up'
    SINGLE_CLICK = 'SingleClick'
    DOUBLE_CLICK = 'DoubleClick'


class KeyboardInteractionType(Enum):
    KEY_PRESS = 'KeyPress'
    KEY_RELEASE = 'KeyRelease'
    INSERT_BY_KEY = 'InsertByKey'


class KeyModifier(Enum):
    CAPS_LOCK = 'CapsLock'
    CONTROL = 'Control'
    SHIFT = 'Shift'
    META = 'Meta'
    ALT = 'Alt'


VIRTUAL_KEYS = {
    '\ue001': ('Cancel', None),
    '\ue002': ('Help', None),
    '\ue003': ('Backspace', None),
    '\ue004': ('Tab', None),
    '\ue005': ('Clear', None),
    '\ue006': ('Return', None),
    '\ue007': ('Enter', None),
    '\ue008': ('Shift', KeyModifier.SHIFT),
    '\ue050': ('Shift', KeyModifier.SHIFT),
    '\ue009': ('Control', KeyModifier.CONTROL),
    '\ue051': ('Control', KeyModifier.CONTROL),
    '\ue00a': ('Alternate', KeyModifier.ALT),
    '\ue052': ('Alternate', KeyModifier.ALT),
    '\ue00b': ('Pause', None),
    '\ue00c': ('Escape', None),
    '\ue00d': ('Space', None),
    '\ue00e': ('PageUp', None),
    '\ue054': ('PageUp', None),
    '\ue00f': ('PageDown', None),
    '\ue055': ('PageDown', None),
    '\ue010': ('End', None),
    '\ue056': ('End', None),
    '\ue011': ('Home', None),
    '\ue057': ('Home', None),
    '\ue012': ('LeftArrow', None),
    '\ue058': ('LeftArrow', None),
    '\ue013': ('UpArrow', None),
    '\ue059': ('UpArrow', None),
    '\ue014': ('RightArrow', None),
    '\ue05a': ('RightArrow', None),
    '\ue015': ('DownArrow', None),
    '\ue05b': ('DownArrow', None),
    '\ue016': ('Insert', None),
    '\ue05c': ('Insert', None),
    '\ue017': ('Delete', None),
    '\ue05d': ('Delete', None),
    '\ue018': ('Semicolon', None),
    '\ue019': ('Equals', None),
    '\ue01a': ('NumberPad0', None),
    '\ue01b': ('NumberPad1', None),
    '\ue01c': ('NumberPad2', None),
    '\ue01d': ('NumberPad3', None),
    '\ue01e': ('NumberPad4', None),
    '\ue01f': ('NumberPad5', None),
    '\ue020': ('NumberPad6', None),
    '\ue021': ('NumberPad7', None),
    '\ue022': ('NumberPad8', None),
    '\ue023': ('NumberPad9', None),
    '\ue024': ('NumberPadMultiply', None),
    '\ue025': ('NumberPadAdd', None),
    '\ue026': ('NumberPadSeparator', None),
    '\ue027': ('NumberPadSubtract', None),
    '\ue028': ('NumberPadDecimal', None),
    '\ue029': ('NumberPadDivide', None),
    '\ue031': ('Function1', None),
    '\ue032': ('Function2', None),
    '\ue033': ('Function3', None),
    '\ue034': ('Function4', None),
    '\ue035': ('Function5', None),
    '\ue036': ('Function6', None),
    '\ue037': ('Function7', None),
    '\ue038': ('Function8', None),
    '\ue039': ('Function9', None),
    '\ue03a': ('Function10', None),
    '\ue03b': ('Function11', None),
    '\ue03c': ('Function12', None),
    '\ue03d': ('Meta', KeyModifier.META),
    '\ue053': ('Meta', KeyModifier.META),
}

MODIFIER_TO_KEY = {
    KeyModifier.SHIFT: 'Shift',
    KeyModifier.CONTROL: 'Control',
    KeyModifier.ALT: 'Alternate',
    KeyModifier.META: 'Meta',
}


@dataclass
class Point:
    x: int
    y: int


@dataclass
class Size:
    width: int
    height: int


@dataclass
class Rect:
    x: int
    y: int
    width: int
    height: int


class AutomationSession:
    def __init__(self, protocol):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        """
        self.protocol = protocol
        self.top_level_handle = ''
        self.current_handle = ''
        self.current_parent_handle = ''
        self.implicit_wait_timeout = 0
        self.page_load_timeout = 3000000
        self.script_timeout = -1

    @property
    def id_(self):
        return self.protocol.id_

    def start_session(self):
        handle = self.protocol.createBrowsingContext()['handle']
        self.switch_to_top_level_browsing_context(handle)

    def stop_session(self):
        self.top_level_handle = ''
        self.current_handle = ''
        self.current_parent_handle = ''
        for handle in self.get_window_handles():
            self.protocol.closeBrowsingContext(handle=handle)

    def create_window(self, type_):
        type_ = type_.capitalize()
        params = {'presentationHint': type_} if type_ else {}
        return self.protocol.createBrowsingContext(**params)['handle']

    def close_window(self):
        if not self.top_level_handle:
            return
        handle = self.top_level_handle
        self.protocol.closeBrowsingContext(handle=handle)

    def maximize_window(self):
        self.protocol.maximizeWindowOfBrowsingContext(handle=self.top_level_handle)

    def hide_window(self):
        self.protocol.hideWindowOfBrowsingContext(handle=self.top_level_handle)

    def get_browsing_context(self):
        return self.protocol.getBrowsingContext(handle=self.top_level_handle)['context']

    def get_window_handles(self):
        contexts = self.protocol.getBrowsingContexts()
        return [c['handle'] for c in contexts['contexts']]

    def set_window_frame(self, x=None, y=None, width=None, height=None):
        params = {}
        if x is not None and y is not None:
            params['origin'] = {'x': x, 'y': y}
        if width is not None and height is not None:
            params['size'] = {'width': width, 'height': height}
        self.protocol.setWindowFrameOfBrowsingContext(handle=self.top_level_handle, **params)

    def add_single_cookie(self, cookie):
        self.protocol.addSingleCookie(browsingContextHandle=self.top_level_handle, cookie=cookie)

    def delete_all_cookies(self):
        self.protocol.deleteAllCookies(browsingContextHandle=self.top_level_handle)

    def delete_single_cookie(self, name):
        self.protocol.deleteSingleCookie(browsingContextHandle=self.top_level_handle, cookieName=name)

    def get_all_cookies(self):
        return self.protocol.getAllCookies(browsingContextHandle=self.top_level_handle)['cookies']

    def execute_script(self, script, args, async_=False):
        parameters = {
            'browsingContextHandle': self.top_level_handle,
            'function': 'function(){\n' + script + '\n}',
            'arguments': list(map(json.dumps, args)),
        }
        if self.current_handle:
            parameters['frameHandle'] = self.current_handle
        if async_:
            parameters['expectsImplicitCallbackArgument'] = True
        if self.script_timeout != -1:
            parameters['callbackTimeout'] = self.script_timeout
        result = self.protocol.evaluateJavaScriptFunction(**parameters, wait_for_response=not async_)
        if async_:
            return result
        else:
            return json.loads(result['result'])

    def evaluate_js_function(self, function, *args, implicit_callback=False, include_frame=True):
        params = {
            'browsingContextHandle': self.top_level_handle,
            'function': function,
            'arguments': list(map(json.dumps, args)),
        }
        if include_frame and self.current_handle:
            params['frameHandle'] = self.current_handle
        if implicit_callback:
            params['expectsImplicitCallbackArgument'] = True
        result = self.protocol.evaluateJavaScriptFunction(**params)
        return json.loads(result['result'])

    def find_elements(self, by, value, single: bool = True, root=None):
        self.wait_for_navigation_to_complete()
        by = by.value if isinstance(by, By) else by
        if by == By.ID.value:
            by = By.CSS_SELECTOR.value
            value = '[id="%s"]' % value
        elif by == By.TAG_NAME.value:
            by = By.CSS_SELECTOR.value
        elif by == By.CLASS_NAME.value:
            by = By.CSS_SELECTOR.value
            value = ".%s" % value
        elif by == By.NAME.value:
            by = By.CSS_SELECTOR.value
            value = '[name="%s"]' % value

        parameters = {
            'browsingContextHandle': self.top_level_handle,
            'function': FIND_NODES,
            'arguments': list(map(json.dumps, [by, root, value, single, self.implicit_wait_timeout])),
            'expectsImplicitCallbackArgument': True,
        }
        if self.current_handle:
            parameters['frameHandle'] = self.current_handle
        if self.implicit_wait_timeout:
            parameters['callbackTimeout'] = self.implicit_wait_timeout + 1000
        result = json.loads(self.protocol.evaluateJavaScriptFunction(**parameters)['result'])
        return result

    def screenshot_as_base64(self, scroll=False, node_id='', clip=True):
        params = {'handle': self.top_level_handle, 'clipToViewport': clip}
        if self.current_handle:
            params['frameHandle'] = self.current_handle
        if scroll:
            params['scrollIntoViewIfNeeded'] = True
        if node_id:
            params['nodeHandle'] = node_id
        return self.protocol.takeScreenshot(**params)['data']

    def switch_to_top_level_browsing_context(self, top_level_handle):
        self.top_level_handle = top_level_handle
        self.current_handle = ''
        self.current_parent_handle = ''

    def switch_to_browsing_context(self, handle):
        self.current_handle = handle
        if not self.current_handle:
            self.current_parent_handle = ''
            return

        resp = self.protocol.resolveParentFrameHandle(browsingContextHandle=self.top_level_handle,
                                                      frameHandle=self.current_handle)
        self.current_parent_handle = resp['result']

    def switch_to_browsing_context_frame(self, context, frame):
        self.protocol.switchToBrowsingContext(browsingContextHandle=context, frameHandle=frame)

    def navigate_broswing_context(self, url):
        self.protocol.navigateBrowsingContext(
            handle=self.top_level_handle, pageLoadTimeout=self.page_load_timeout, url=url
        )

    def go_back_in_browsing_context(self):
        self.protocol.goBackInBrowsingContext(
            handle=self.top_level_handle, pageLoadTimeout=self.page_load_timeout
        )

    def go_forward_in_browsing_context(self):
        self.protocol.goForwardInBrowsingContext(
            handle=self.top_level_handle, pageLoadTimeout=self.page_load_timeout
        )

    def reload_browsing_context(self):
        self.protocol.reloadBrowsingContext(
            handle=self.top_level_handle, pageLoadTimeout=self.page_load_timeout
        )

    def switch_to_frame(self, frame_ordinal=None, frame_handle=None):
        params = {'browsingContextHandle': self.top_level_handle}
        if self.current_handle:
            params['frameHandle'] = self.current_handle
        if frame_ordinal is not None:
            params['ordinal'] = frame_ordinal
        elif frame_handle is not None:
            params['nodeHandle'] = frame_handle.node_id
        resp = self.protocol.resolveChildFrameHandle(**params)['result']
        self.switch_to_browsing_context_frame(self.top_level_handle, resp)
        self.switch_to_browsing_context(resp)

    def switch_to_window(self, handle):
        self.switch_to_browsing_context_frame(handle, '')
        self.switch_to_top_level_browsing_context(handle)

    def perform_keyboard_interactions(self, interactions):
        for interaction in interactions:
            type_ = interaction['type']
            interaction['type'] = type_.value if isinstance(type_, KeyboardInteractionType) else type_
        self.protocol.performKeyboardInteractions(handle=self.top_level_handle, interactions=interactions)

    def perform_mouse_interaction(self, x, y, button: MouseButton, interaction: MouseInteraction, modifiers=None):
        modifiers = [] if modifiers is None else modifiers
        self.protocol.performMouseInteraction(handle=self.top_level_handle, position={'x': x, 'y': y},
                                              button=button.value, interaction=interaction.value, modifiers=modifiers)

    def perform_interaction_sequence(self, sources, steps):
        params = {
            'handle': self.top_level_handle,
            'inputSources': sources,
            'steps': steps,
        }
        if self.current_handle:
            params['frameHandle'] = self.current_handle
        self.protocol.performInteractionSequence(**params)

    def wait_for_navigation_to_complete(self):
        params = {'browsingContextHandle': self.top_level_handle, 'pageLoadTimeout': self.page_load_timeout}
        if self.current_handle:
            params['frameHandle'] = self.current_handle
        self.protocol.waitForNavigationToComplete(**params)

    def accept_current_javascript_dialog(self):
        self.protocol.acceptCurrentJavaScriptDialog(browsingContextHandle=self.top_level_handle)

    def dismiss_current_javascript_dialog(self):
        self.protocol.dismissCurrentJavaScriptDialog(browsingContextHandle=self.top_level_handle)

    def set_user_input_for_current_javascript_prompt(self, user_input):
        self.protocol.setUserInputForCurrentJavaScriptPrompt(
            browsingContextHandle=self.top_level_handle, userInput=user_input
        )

    def message_of_current_javascript_dialog(self):
        return self.protocol.messageOfCurrentJavaScriptDialog(browsingContextHandle=self.top_level_handle)['message']

    def compute_element_layout(self, node_id, scroll_if_needed, coordinate_system):
        return self.protocol.computeElementLayout(
            browsingContextHandle=self.top_level_handle, nodeHandle=node_id,
            scrollIntoViewIfNeeded=scroll_if_needed, coordinateSystem=coordinate_system,
            frameHandle='' if self.current_handle is None else self.current_handle
        )

    def select_option_element(self, node_id):
        self.protocol.selectOptionElement(
            browsingContextHandle=self.top_level_handle, nodeHandle=node_id,
            frameHandle='' if self.current_handle is None else self.current_handle
        )



import dataclasses
import logging
import socket
import time
from typing import List, Mapping

from zeroconf import InterfaceChoice, InterfacesType, IPVersion, ServiceBrowser, ServiceListener, Zeroconf

from pymobiledevice3.lockdown import LockdownClient, create_using_tcp

SERVICE_NAME = '_apple-mobdev2._tcp.local.'

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class BonjourDevice:
    name: str
    mac_address: str
    ipv4: List[str]
    ipv6: List[str]
    lockdown: LockdownClient

    def asdict(self) -> Mapping:
        return {
            'name': self.name,
            'mac_address': self.mac_address,
            'ipv4': self.ipv4,
            'ipv6': self.ipv6,
            'lockdown_info': self.lockdown.all_values
        }


class BonjourListener(ServiceListener):
    def __init__(self, pair_records: List[Mapping] = None):
        super().__init__()
        self.pair_records = [] if pair_records is None else pair_records
        self.discovered_devices = {}

    def update_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        logger.debug(f'Service {name} updated')

    def remove_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        logger.debug(f'Service {name} removed')
        self.discovered_devices.pop(name)

    def add_service(self, zc: Zeroconf, type_: str, name: str) -> None:
        info = zc.get_service_info(type_, name)
        logger.debug(f'Service {name} added, service info: {info}')

        ipv4 = [socket.inet_ntop(socket.AF_INET, address) for address in info.addresses_by_version(IPVersion.V4Only)]
        ipv6 = [socket.inet_ntop(socket.AF_INET6, address) for address in info.addresses_by_version(IPVersion.V6Only)]

        try:
            lockdown = create_using_tcp(hostname=ipv4[0], autopair=False)

            for pair_record in self.pair_records:
                lockdown = create_using_tcp(hostname=ipv4[0], autopair=False, pair_record=pair_record)
                if lockdown.paired:
                    break
        except ConnectionRefusedError:
            logger.debug('Service failed to establish a lockdown connection')
            return

        self.discovered_devices[name] = BonjourDevice(name=name, mac_address=name.split('@')[0], ipv4=ipv4, ipv6=ipv6,
                                                      lockdown=lockdown)


def browse(timeout: int, interfaces: InterfacesType = InterfaceChoice.All, pair_records: List[Mapping] = None) -> \
        Mapping[str, BonjourDevice]:
    with Zeroconf(interfaces=interfaces) as zc:
        listener = BonjourListener(pair_records=pair_records)
        ServiceBrowser(zc, SERVICE_NAME, listener)
        time.sleep(timeout)
        return listener.discovered_devices



from dataclasses import asdict, dataclass
from typing import List

from pymobiledevice3.services.web_protocol.automation_session import RESOURCES, Point, Rect, Size
from pymobiledevice3.services.web_protocol.element import WebElement
from pymobiledevice3.services.web_protocol.selenium_api import By, SeleniumApi
from pymobiledevice3.services.web_protocol.switch_to import SwitchTo

ENTER_FULLSCREEN = (RESOURCES / 'enter_fullscreen.js').read_text()


@dataclass
class Cookie:
    name: str
    value: str
    domain: str = ''
    path: str = '/'
    expires: int = 0
    httpOnly: bool = False
    secure: bool = False
    session: bool = True
    sameSite: str = 'None'

    @classmethod
    def from_automation(cls, d):
        d.pop('size')
        return cls(**d)


class WebDriver(SeleniumApi):
    def __init__(self, session):
        """
        :param pymobiledevice3.services.web_protocol.automation_session.AutomationSession session: Automation session.
        """
        self.session = session
        self.switch_to = SwitchTo(session)

    def add_cookie(self, cookie: Cookie):
        """ Adds a cookie to your current session. """
        if isinstance(cookie, Cookie):
            cookie = asdict(cookie)
        self.session.add_single_cookie(cookie)

    def back(self):
        """ Goes one step backward in the browser history. """
        self.session.wait_for_navigation_to_complete()
        self.session.go_back_in_browsing_context()
        self.session.switch_to_browsing_context('')

    def close(self):
        """ Closes the current window. """
        self.session.close_window()

    @property
    def current_url(self):
        """ Gets the URL of the current page. """
        self.session.wait_for_navigation_to_complete()
        return self.session.get_browsing_context()['url']

    @property
    def current_window_handle(self):
        """ Returns the handle of the current window. """
        return self.session.get_browsing_context()['handle']

    def delete_all_cookies(self):
        """ Delete all cookies in the scope of the session. """
        self.session.delete_all_cookies()

    def delete_cookie(self, name: str):
        """ Deletes a single cookie with the given name. """
        self.session.delete_single_cookie(name)

    def execute_async_script(self, script: str, *args):
        """
        Asynchronously Executes JavaScript in the current window/frame.
        :param script: The JavaScript to execute.
        :param args: Any applicable arguments for your JavaScript.
        """
        return self.session.execute_script(script, args, async_=True)

    def execute_script(self, script: str, *args):
        """
        Synchronously Executes JavaScript in the current window/frame.
        :param script: The JavaScript to execute.
        :param args: Any applicable arguments for your JavaScript.
        """
        return self.session.execute_script(script, args)

    def find_element(self, by=By.ID, value=None) -> WebElement:
        """ Find an element given a By strategy and locator. """
        elem = self.session.find_elements(by, value)
        return None if elem is None else WebElement(self.session, elem)

    def find_elements(self, by=By.ID, value=None) -> List[WebElement]:
        """ Find elements given a By strategy and locator. """
        elements = self.session.find_elements(by, value, single=False)
        return list(map(lambda elem: WebElement(self.session, elem), elements))

    def forward(self):
        """ Goes one step forward in the browser history. """
        self.session.wait_for_navigation_to_complete()
        self.session.go_forward_in_browsing_context()
        self.session.switch_to_browsing_context('')

    def fullscreen_window(self):
        """ Invokes the window manager-specific 'full screen' operation. """
        self.session.evaluate_js_function(ENTER_FULLSCREEN, implicit_callback=True)

    def get(self, url: str):
        """ Loads a web page in the current browser session. """
        self.session.wait_for_navigation_to_complete()
        self.session.navigate_broswing_context(url)
        self.session.switch_to_browsing_context('')

    def get_cookie(self, name: str) -> Cookie:
        """ Get a single cookie by name. Returns the cookie if found, None if not. """
        for cookie in self.get_cookies():
            if cookie.name == name:
                return cookie

    def get_cookies(self) -> List[Cookie]:
        """ Returns cookies visible in the current session. """
        return list(map(Cookie.from_automation, self.session.get_all_cookies()))

    @property
    def screenshot_as_base64(self) -> str:
        """ Gets the screenshot of the current window as a base64 encoded string. """
        return self.session.screenshot_as_base64()

    def get_window_position(self) -> Point:
        """ Gets the x,y position of the current window. """
        rect = self.get_window_rect()
        return Point(x=rect.x, y=rect.y)

    def get_window_rect(self) -> Rect:
        """ Gets the x, y coordinates of the window as well as height and width of the current window. """
        context = self.session.get_browsing_context()
        return Rect(
            context['windowOrigin']['x'], context['windowOrigin']['y'],
            context['windowSize']['width'], context['windowSize']['height']
        )

    def get_window_size(self) -> Size:
        """ Gets the width and height of the current window. """
        rect = self.get_window_rect()
        return Size(height=rect.height, width=rect.width)

    def implicitly_wait(self, time_to_wait):
        """ Sets a sticky timeout to implicitly wait for an element to be found, or a command to complete. """
        self.session.implicit_wait_timeout = time_to_wait * 1000

    def maximize_window(self):
        """ Maximizes the current window. """
        self.session.maximize_window()

    def minimize_window(self):
        """ Invokes the window manager-specific 'minimize' operation. """
        self.session.hide_window()

    @property
    def page_source(self) -> str:
        """ Gets the source of the current page. """
        return self.session.evaluate_js_function('function() { return document.documentElement.outerHTML; }')

    def refresh(self):
        """ Refreshes the current page. """
        self.session.wait_for_navigation_to_complete()
        self.session.reload_browsing_context()
        self.session.switch_to_browsing_context('')

    def set_window_position(self, x, y):
        """ Sets the x,y position of the current window. """
        self.set_window_rect(x=int(x, 0), y=int(y, 0))

    def set_window_rect(self, x=None, y=None, width=None, height=None):
        """ Sets the x, y coordinates of the window as well as height and width of the current window. """
        self.session.set_window_frame(x, y, width, height)

    def set_window_size(self, width, height):
        """ Sets the width and height of the current window. """
        self.set_window_rect(width=int(width, 0), height=int(height, 0))

    def start_session(self):
        """ Creates a new session. """
        self.session.start_session()

    @property
    def title(self) -> str:
        """ Returns the title of the current page. """
        self.session.wait_for_navigation_to_complete()
        return self.session.evaluate_js_function('function() { return document.title; }')

    @property
    def window_handles(self) -> List[str]:
        """ Returns the handles of all windows within the current session. """
        return self.session.get_window_handles()



from datetime import datetime
from typing import IO

import click
from pygments import formatters, highlight, lexers

from pymobiledevice3.cli.cli_common import Command, print_hex
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.pcapd import PcapdService


@click.group()
def cli():
    """ apps cli """
    pass


def print_packet_header(packet, color: bool):
    date = datetime.fromtimestamp(packet.seconds + (packet.microseconds / 1000000))
    data = (
        f'{date}: '
        f'Process {packet.comm} ({packet.pid}), '
        f'Interface: {packet.interface_name} ({packet.interface_type.name}), '
        f'Family: {packet.protocol_family.name}'
    )
    if not color:
        print(data)
    else:
        print(highlight(data, lexers.HspecLexer(), formatters.TerminalTrueColorFormatter(style='native')), end='')


def print_packet(packet, color: bool):
    """ Return the packet so it can be chained in a generator """
    print_packet_header(packet, color)
    print_hex(packet.data, color)
    return packet


@cli.command(cls=Command)
@click.argument('out', type=click.File('wb'), required=False)
@click.option('-c', '--count', type=click.INT, default=-1, help='Number of packets to sniff. Omit to endless sniff.')
@click.option('--process', default=None, help='Process to filter. Omit for all.')
@click.option('--color/--no-color', default=True)
def pcap(lockdown: LockdownClient, out: IO, count: int, process: str, color: bool):
    """ sniff device traffic """
    service = PcapdService(lockdown=lockdown)
    packets_generator = service.watch(packets_count=count, process=process)

    if out is not None:
        packets_generator_with_print = map(lambda p: print_packet(p, color), packets_generator)
        service.write_to_pcap(out, packets_generator_with_print)
        return

    for packet in packets_generator:
        print_packet(packet, color)



import asyncio
import logging
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from functools import update_wrapper
from typing import Optional, Type

import click
import inquirer
import IPython
import uvicorn
from inquirer.themes import GreenPassion
from prompt_toolkit import HTML, PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import FileHistory
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.styles import style_from_pygments_cls
from pygments import formatters, highlight, lexers
from pygments.styles import get_style_by_name

from pymobiledevice3.cli.cli_common import Command, wait_return
from pymobiledevice3.common import get_home_folder
from pymobiledevice3.exceptions import InspectorEvaluateError, LaunchingApplicationError, \
    RemoteAutomationNotEnabledError, WebInspectorNotEnabledError, WirError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.web_protocol.cdp_server import app
from pymobiledevice3.services.web_protocol.driver import By, Cookie, WebDriver
from pymobiledevice3.services.web_protocol.inspector_session import InspectorSession
from pymobiledevice3.services.webinspector import SAFARI, Page, WebinspectorService

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ webinspector cli """
    pass


@cli.group()
def webinspector():
    """ webinspector options """
    pass


def catch_errors(func):
    def catch_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except LaunchingApplicationError:
            logger.error('Unable to launch application (try to unlock device)')
        except WebInspectorNotEnabledError:
            logger.error('Web inspector is not enable')
        except RemoteAutomationNotEnabledError:
            logger.error('Remote automation is not enable')

    return update_wrapper(catch_function, func)


def reload_pages(inspector: WebinspectorService):
    inspector.get_open_pages()
    # Best effort.
    inspector.flush_input(2)


def create_webinspector_and_launch_app(lockdown: LockdownClient, timeout: float, app: str):
    inspector = WebinspectorService(lockdown=lockdown)
    inspector.connect(timeout)
    application = inspector.open_app(app)
    return inspector, application


@webinspector.command(cls=Command)
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--timeout', default=3, show_default=True, type=float)
@catch_errors
def opened_tabs(lockdown: LockdownClient, verbose, timeout):
    """
    Show all currently opened tabs.

    \b
    Opt-in:
        Settings -> Safari -> Advanced -> Web Inspector
    """
    inspector = WebinspectorService(lockdown=lockdown, loop=asyncio.get_event_loop())
    inspector.connect(timeout)
    while not inspector.connected_application:
        inspector.flush_input()
    reload_pages(inspector)
    for app_id, app_ in inspector.connected_application.items():
        if app_id not in inspector.application_pages:
            continue
        if verbose:
            print(f'{app_.name}    id: {app_id}')
        else:
            print(app_.name)
        for page_id, page in inspector.application_pages[app_id].items():
            if verbose:
                print(f' - {page.web_url}    id: {page_id}')
            else:
                print(f' - {page.web_url}')
    inspector.close()


@webinspector.command(cls=Command)
@click.argument('url')
@click.option('-t', '--timeout', default=3, show_default=True, type=float)
@catch_errors
def launch(lockdown: LockdownClient, url, timeout):
    """
    Launch a specific URL in Safari.

    \b
    Opt-in:
        Settings -> Safari -> Advanced -> Web Inspector
        Settings -> Safari -> Advanced -> Remote Automation
    """
    inspector, safari = create_webinspector_and_launch_app(lockdown, timeout, SAFARI)
    session = inspector.automation_session(safari)
    driver = WebDriver(session)
    print('Starting session')
    driver.start_session()
    print('Getting URL')
    driver.get(url)
    wait_return()
    session.stop_session()
    inspector.close()


SHELL_USAGE = '''
# This shell allows you to control the web with selenium like API.
# The first thing you should do is creating a session:
driver.start_session()

# Then, you can navigate by using the get method:
driver.get("https://google.com")

# You can search for a specific element using the `By` enum:
driver.find_element(By.TAG_NAME, 'input')

# You can also add cookies:
driver.add_cookie(
    Cookie(name='tz', value='UTC', domain='.github.com', path='/', expires=0, httpOnly=False, secure=True,
    session=True, sameSite='None')
)

# See selenium api for more features.
'''


@webinspector.command(cls=Command)
@click.option('-t', '--timeout', default=3, show_default=True, type=float)
@catch_errors
def shell(lockdown: LockdownClient, timeout):
    """
    Create an IPython shell for interacting with a WebView.

    \b
    Opt-in:
        Settings -> Safari -> Advanced -> Web Inspector
        Settings -> Safari -> Advanced -> Remote Automation
    """
    inspector, safari = create_webinspector_and_launch_app(lockdown, timeout, SAFARI)
    session = inspector.automation_session(safari)
    driver = WebDriver(session)
    try:
        IPython.embed(
            header=highlight(SHELL_USAGE, lexers.PythonLexer(), formatters.TerminalTrueColorFormatter(style='native')),
            user_ns={
                'driver': driver,
                'Cookie': Cookie,
                'By': By,
            })
    finally:
        session.stop_session()
        inspector.close()


@webinspector.command(cls=Command)
@click.option('-t', '--timeout', default=3, show_default=True, type=float)
@click.option('--automation', is_flag=True, help='Use remote automation')
@click.argument('url', required=False, default='')
@catch_errors
def js_shell(lockdown: LockdownClient, timeout, automation, url):
    """
    Create a javascript shell. This interpreter runs on your local machine,
    but evaluates each expression on the remote

    \b
    Opt-in:
        Settings -> Safari -> Advanced -> Web Inspector

    \b
    for automation also enable:
        Settings -> Safari -> Advanced -> Remote Automation
    """

    js_shell_class = AutomationJsShell if automation else InspectorJsShell
    asyncio.run(run_js_shell(js_shell_class, lockdown, timeout, url))


udid = ''


def create_app():
    inspector = WebinspectorService(lockdown=LockdownClient(udid))
    app.state.inspector = inspector
    return app


@webinspector.command(cls=Command)
@click.option('--host', default='127.0.0.1')
@click.option('--port', type=click.INT, default=9222)
def cdp(lockdown: LockdownClient, host, port):
    """
    Start a CDP server for debugging WebViews.

    \b
    In order to debug the WebView that way, open in Google Chrome:
        chrome://inspect/#devices
    """
    global udid
    udid = lockdown.udid
    uvicorn.run('pymobiledevice3.cli.webinspector:create_app', host=host, port=port, factory=True,
                ws_ping_timeout=None, ws='wsproto', loop='asyncio')


class JsShell(ABC):
    def __init__(self):
        super().__init__()
        self.prompt_session = PromptSession(lexer=PygmentsLexer(lexers.JavascriptLexer),
                                            auto_suggest=AutoSuggestFromHistory(),
                                            style=style_from_pygments_cls(get_style_by_name('stata-dark')),
                                            history=FileHistory(self.webinspector_history_path()))

    @classmethod
    @abstractmethod
    def create(cls, lockdown: LockdownClient, timeout: float, app: str):
        pass

    @abstractmethod
    async def evaluate_expression(self, exp):
        pass

    @abstractmethod
    async def navigate(self, url: str):
        pass

    async def js_iter(self):
        with patch_stdout(True):
            exp = await self.prompt_session.prompt_async(HTML('<style fg="cyan"><b>&gt;</b></style> '))

        if not exp.strip():
            return

        result = await self.evaluate_expression(exp)
        colorful_result = highlight(f'{result}', lexers.JavascriptLexer(),
                                    formatters.TerminalTrueColorFormatter(style='stata-dark'))
        print(colorful_result, end='')

    async def start(self, url: str = ''):
        if url:
            await self.navigate(url)
        while True:
            try:
                await self.js_iter()
            except WirError as e:
                logger.error(e)
            except (KeyboardInterrupt, InspectorEvaluateError):  # KeyboardInterrupt Control-C
                pass
            except EOFError:  # Control-D
                return

    @staticmethod
    def webinspector_history_path() -> str:
        return str(get_home_folder() / 'webinspector_history')


class AutomationJsShell(JsShell):
    def __init__(self, driver: WebDriver):
        super().__init__()
        self.driver = driver

    @classmethod
    @asynccontextmanager
    async def create(cls, lockdown: LockdownClient, timeout: float, app: str) -> 'AutomationJsShell':
        inspector, application = create_webinspector_and_launch_app(lockdown, timeout, app)
        automation_session = inspector.automation_session(application)
        driver = WebDriver(automation_session)
        driver.start_session()
        try:
            yield cls(driver)
        finally:
            automation_session.stop_session()
            inspector.close()

    async def evaluate_expression(self, exp: str):
        return self.driver.execute_script(f'return {exp}')

    async def navigate(self, url: str):
        self.driver.get(url)


class InspectorJsShell(JsShell):
    def __init__(self, inspector_session: InspectorSession):
        super().__init__()
        self.inspector_session = inspector_session

    @classmethod
    @asynccontextmanager
    async def create(cls, lockdown: LockdownClient, timeout: float, app: str) -> 'InspectorJsShell':
        inspector, application = create_webinspector_and_launch_app(lockdown, timeout, app)
        page = InspectorJsShell.query_page(inspector)
        if page is None:
            raise click.exceptions.Exit()

        inspector_session = await inspector.inspector_session(application, page)
        await inspector_session.runtime_enable()
        try:
            yield cls(inspector_session)
        finally:
            inspector.close()

    async def evaluate_expression(self, exp: str):
        return await self.inspector_session.runtime_evaluate(exp)

    async def navigate(self, url: str):
        await self.inspector_session.navigate_to_url(url)

    @staticmethod
    def query_page(inspector: WebinspectorService) -> Optional[Page]:
        reload_pages(inspector)
        available_pages = list(inspector.get_open_pages().get('Safari', []))
        if not available_pages:
            logger.error('Unable to find available pages (try to unlock device)')
            return

        page_query = [inquirer.List('page', message='choose page', choices=available_pages, carousel=True)]
        page = inquirer.prompt(page_query, theme=GreenPassion(), raise_keyboard_interrupt=True)['page']
        return page


async def run_js_shell(js_shell_class: Type[JsShell], lockdown: LockdownClient,
                       timeout: float, url: str):
    async with js_shell_class.create(lockdown, timeout, SAFARI) as js_shell_instance:
        await js_shell_instance.start(url)



import json
import logging
from functools import update_wrapper
from pathlib import Path
from typing import List
from urllib.error import URLError
from urllib.request import urlopen

import click
import requests
from tqdm import tqdm

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.common import get_home_folder
from pymobiledevice3.exceptions import AlreadyMountedError, NotMountedError, UnsupportedCommandError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.mobile_image_mounter import DeveloperDiskImageMounter, MobileImageMounterService, \
    PersonalizedImageMounter

DISK_IMAGE_TREE = 'https://api.github.com/repos/pdso/DeveloperDiskImage/git/trees/master'
DEVELOPER_DISK_IMAGE_URL = 'https://github.com/pdso/DeveloperDiskImage/raw/master/{ios_version}/{file_name}'

logger = logging.getLogger(__name__)


def catch_errors(func):
    def catch_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AlreadyMountedError:
            logger.error('Given image was already mounted')
        except UnsupportedCommandError:
            logger.error('Your iOS version doesn\'t support this command')

    return update_wrapper(catch_function, func)


@click.group()
def cli():
    """ mounter cli """
    pass


@cli.group()
def mounter():
    """ mounter options """
    pass


@mounter.command('list', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_list(lockdown: LockdownClient, color):
    """ list all mounted images """
    output = []

    images = MobileImageMounterService(lockdown=lockdown).copy_devices()
    for image in images:
        image_signature = image.get('ImageSignature')
        if image_signature is not None:
            image['ImageSignature'] = image_signature.hex()
        output.append(image)

    print_json(output, colored=color)


@mounter.command('lookup', cls=Command)
@click.option('--color/--no-color', default=True)
@click.argument('image_type')
def mounter_lookup(lockdown: LockdownClient, color, image_type):
    """ lookup mounter image type """
    try:
        signature = MobileImageMounterService(lockdown=lockdown).lookup_image(image_type)
        print_json(signature, colored=color)
    except NotMountedError:
        logger.error(f'Disk image of type: {image_type} is not mounted')


@mounter.command('umount-developer', cls=Command)
@catch_errors
def mounter_umount_developer(lockdown: LockdownClient):
    """ unmount Developer image """
    try:
        DeveloperDiskImageMounter(lockdown=lockdown).umount()
        logger.info('Developer image unmounted successfully')
    except NotMountedError:
        logger.error('Developer image isn\'t currently mounted')


@mounter.command('umount-personalized', cls=Command)
@catch_errors
def mounter_umount_personalized(lockdown: LockdownClient):
    """ unmount Personalized image """
    try:
        PersonalizedImageMounter(lockdown=lockdown).umount()
        logger.info('Personalized image unmounted successfully')
    except NotMountedError:
        logger.error('Personalized image isn\'t currently mounted')


def download_file(url, local_filename):
    logger.debug(f'downloading: {local_filename}')
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        total_size_in_bytes = int(r.headers.get('content-length', 0))

        with tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True, dynamic_ncols=True) as progress_bar:
            with open(local_filename, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    progress_bar.update(len(chunk))
                    f.write(chunk)

    return local_filename


def get_all_versions() -> List[str]:
    data = urlopen(DISK_IMAGE_TREE).read()
    json_data = json.loads(data)
    return [item.get('path') for item in json_data.get('tree')][0:-3]


@mounter.command('mount-developer', cls=Command)
@click.argument('image', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('signature', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@catch_errors
def mounter_mount_developer(lockdown: LockdownClient, image: str, signature: str):
    """ mount developer image """
    DeveloperDiskImageMounter(lockdown=lockdown).mount(Path(image), Path(signature))
    logger.info('Developer image mounted successfully')


@mounter.command('mount-personalized', cls=Command)
@click.argument('image', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('trust-cache', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('build-manifest', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@catch_errors
def mounter_mount_personalized(lockdown: LockdownClient, image: str, trust_cache: str, build_manifest: str):
    """ mount personalized image """
    PersonalizedImageMounter(lockdown=lockdown).mount(Path(image), Path(build_manifest), Path(trust_cache))
    logger.info('Personalized image mounted successfully')


@mounter.command('auto-mount', cls=Command)
@click.option('-x', '--xcode', type=click.Path(exists=True, dir_okay=True, file_okay=False),
              help='Xcode application path used to figure out automatically the DeveloperDiskImage path')
@click.option('-v', '--version', help='use a different DeveloperDiskImage version from the one retrieved by lockdown'
                                      'connection')
def mounter_auto_mount(lockdown: LockdownClient, xcode: str, version: str):
    """ auto-detect correct DeveloperDiskImage and mount it """
    image_type = 'Developer'

    if xcode is None:
        # avoid "default"-ing this option, because Windows and Linux won't have this path
        xcode = Path('/Applications/Xcode.app')
        if not (xcode.exists()):
            xcode = get_home_folder() / 'Xcode.app'
            xcode.mkdir(parents=True, exist_ok=True)

    image_mounter = DeveloperDiskImageMounter(lockdown=lockdown)
    if image_mounter.is_image_mounted(image_type):
        logger.error('DeveloperDiskImage is already mounted')
        return

    logger.debug('trying to figure out the best suited DeveloperDiskImage')
    if version is None:
        version = lockdown.sanitized_ios_version
    image_dir = f'{xcode}/Contents/Developer/Platforms/iPhoneOS.platform/DeviceSupport/{version}'
    image_path = f'{image_dir}/DeveloperDiskImage.dmg'
    signature = f'{image_path}.signature'
    developer_disk_image_dir = Path(image_path).parent

    image_path = Path(image_path)
    signature = Path(signature)

    if not image_path.exists():
        try:
            available_versions = get_all_versions()
            if version not in available_versions:
                logger.error(
                    f'Unable to find DeveloperDiskImage for {version}. available versions: {available_versions}')
                return
        except URLError:
            logger.warning('failed to query DeveloperDiskImage versions')

    try:
        developer_disk_image_dir.mkdir(exist_ok=True)

        if not image_path.exists():
            download_file(DEVELOPER_DISK_IMAGE_URL.format(ios_version=version, file_name=image_path.name), image_path)

        if not signature.exists():
            download_file(DEVELOPER_DISK_IMAGE_URL.format(ios_version=version, file_name=signature.name), signature)

    except PermissionError:
        logger.error(
            f'DeveloperDiskImage could not be saved to Xcode default path ({developer_disk_image_dir}). '
            f'Please make sure your user has the necessary permissions')
        return

    image_mounter.mount(image_path, signature)
    logger.info('DeveloperDiskImage mounted successfully')


@mounter.command('query-developer-mode-status', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_developer_mode_status(lockdown: LockdownClient, color):
    """ Query developer mode status """
    print_json(MobileImageMounterService(lockdown=lockdown).query_developer_mode_status(), colored=color)


@mounter.command('query-nonce', cls=Command)
@click.option('--image-type')
@click.option('--color/--no-color', default=True)
def mounter_query_nonce(lockdown: LockdownClient, image_type: str, color: bool):
    """ Query nonce """
    print_json(MobileImageMounterService(lockdown=lockdown).query_nonce(image_type), colored=color)


@mounter.command('query-personalization-identifiers', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_personalization_identifiers(lockdown: LockdownClient, color):
    """ Query personalization identifiers """
    print_json(MobileImageMounterService(lockdown=lockdown).query_personalization_identifiers(), colored=color)


@mounter.command('query-personalization-manifest', cls=Command)
@click.option('--color/--no-color', default=True)
def mounter_query_personalization_manifest(lockdown: LockdownClient, color):
    """ Query personalization manifest """
    result = []
    mounter = MobileImageMounterService(lockdown=lockdown)
    for device in mounter.copy_devices():
        result.append(mounter.query_personalization_manifest(device['PersonalizedImageType'], device['ImageSignature']))
    print_json(result, colored=color)


@mounter.command('roll-personalization-nonce', cls=Command)
def mounter_roll_personalization_nonce(lockdown: LockdownClient):
    MobileImageMounterService(lockdown=lockdown).roll_personalization_nonce()


@mounter.command('roll-cryptex-nonce', cls=Command)
def mounter_roll_cryptex_nonce(lockdown: LockdownClient):
    """ Roll cryptex nonce (will reboot) """
    MobileImageMounterService(lockdown=lockdown).roll_cryptex_nonce()



import plistlib
from pathlib import Path

import click

from pymobiledevice3.bonjour import browse
from pymobiledevice3.cli.cli_common import print_json

DEFAULT_BROWSE_TIMEOUT = 5


@click.group()
def cli():
    """ bonjour cli """
    pass


@cli.group('bonjour')
def bonjour_cli():
    """ bonjour options """
    pass


@bonjour_cli.command('browse')
@click.option('--timeout', default=DEFAULT_BROWSE_TIMEOUT, type=click.INT)
@click.option('--pair-records', type=click.Path(dir_okay=True, file_okay=False, exists=True),
              help='pair records to attempt validation with')
@click.option('--color/--no-color', default=True)
def cli_browse(timeout: int, pair_records: str, color: bool):
    """ browse devices over bonjour """
    records = []
    if pair_records is not None:
        for record in Path(pair_records).glob('*.plist'):
            records.append(plistlib.loads(record.read_bytes()))

    output = []
    for device in browse(timeout, pair_records=records).values():
        device = device.asdict()
        output.append(device)

    print_json(output, colored=color)



import os
import posixpath
from typing import Callable, List, Mapping

from pymobiledevice3.exceptions import AppInstallError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import AfcService
from pymobiledevice3.services.base_service import BaseService

GET_APPS_ADDITIONAL_INFO = {'ReturnAttributes': ['CFBundleIdentifier', 'StaticDiskUsage', 'DynamicDiskUsage']}


class InstallationProxyService(BaseService):
    SERVICE_NAME = 'com.apple.mobile.installation_proxy'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def _watch_completion(self, handler: Callable = None, *args) -> None:
        while True:
            response = self.service.recv_plist()
            if not response:
                break
            error = response.get('Error')
            if error:
                raise AppInstallError(f'{error}: {response.get("ErrorDescription")}')
            completion = response.get('PercentComplete')
            if completion:
                if handler:
                    self.logger.debug('calling handler')
                    handler(completion, *args)
                self.logger.info(f'{response.get("PercentComplete")}% Complete')
            if response.get('Status') == 'Complete':
                return
        raise AppInstallError()

    def send_cmd_for_bundle_identifier(self, bundle_identifier: str, cmd: str = 'Archive', options: Mapping = None,
                                       handler: Mapping = None, *args) -> None:
        """ send a low-level command to installation relay """
        cmd = {'Command': cmd,
               'ApplicationIdentifier': bundle_identifier}

        if options is None:
            options = {}

        cmd.update({'ClientOptions': options})
        self.service.send_plist(cmd)
        self._watch_completion(handler, *args)

    def install(self, ipa_path: str, options: Mapping = None, handler: Callable = None, *args) -> None:
        """ install given ipa from device path """
        self.install_from_local(ipa_path, 'Install', options, handler, args)

    def upgrade(self, ipa_path: str, options: Mapping = None, handler: Callable = None, *args) -> None:
        """ upgrade given ipa from device path """
        self.install_from_local(ipa_path, 'Upgrade', options, handler, args)

    def restore(self, bundle_identifier: str, options: Mapping = None, handler: Callable = None, *args) -> None:
        """ no longer supported on newer iOS versions """
        self.send_cmd_for_bundle_identifier(bundle_identifier, 'Restore', options, handler, args)

    def uninstall(self, bundle_identifier: str, options: Mapping = None, handler: Callable = None, *args) -> None:
        """ uninstall given bundle_identifier """
        self.send_cmd_for_bundle_identifier(bundle_identifier, 'Uninstall', options, handler, args)

    def install_from_local(self, ipa_path: str, cmd='Install', options: Mapping = None, handler: Callable = None,
                           *args) -> None:
        """ upload given ipa onto device and install it """
        if options is None:
            options = {}
        remote_path = posixpath.join('/', os.path.basename(ipa_path))
        with AfcService(self.lockdown) as afc:
            afc.set_file_contents(remote_path, open(ipa_path, 'rb').read())
        cmd = {'Command': cmd,
               'ClientOptions': options,
               'PackagePath': remote_path}
        self.service.send_plist(cmd)
        self._watch_completion(handler, args)

    def check_capabilities_match(self, capabilities: Mapping = None, options: Mapping = None) -> Mapping:
        if options is None:
            options = {}
        cmd = {'Command': 'CheckCapabilitiesMatch',
               'ClientOptions': options}

        if capabilities:
            cmd['Capabilities'] = capabilities

        return self.service.send_recv_plist(cmd).get('LookupResult')

    def browse(self, options: Mapping = None, attributes: List[str] = None) -> List[Mapping]:
        if options is None:
            options = {}
        if attributes:
            options['ReturnAttributes'] = attributes

        cmd = {'Command': 'Browse',
               'ClientOptions': options}

        self.service.send_plist(cmd)

        result = []
        while True:
            response = self.service.recv_plist()
            if not response:
                break

            data = response.get('CurrentList')
            if data is not None:
                result += data

            if response.get('Status') == 'Complete':
                break

        return result

    def lookup(self, options: Mapping = None) -> Mapping:
        """ search installation database """
        if options is None:
            options = {}
        cmd = {'Command': 'Lookup', 'ClientOptions': options}
        return self.service.send_recv_plist(cmd).get('LookupResult')

    def get_apps(self, app_types: List[str] = None) -> Mapping[str, Mapping]:
        """ get applications according to given criteria """
        result = self.lookup()
        # query for additional info
        additional_info = self.lookup(GET_APPS_ADDITIONAL_INFO)
        for bundle_identifier, app in additional_info.items():
            result[bundle_identifier].update(app)
        # filter results
        filtered_result = {}
        for bundle_identifier, app in result.items():
            if (app_types is None) or (app['ApplicationType'] in app_types):
                filtered_result[bundle_identifier] = app
        return filtered_result



import typing
from dataclasses import dataclass
from enum import Enum

from packaging.version import Version

from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.remote_server import MessageAux, RemoteServer


class SerializedObject:
    def __init__(self, fields: typing.Mapping):
        self._fields = fields


class AXAuditInspectorFocus_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)

    @property
    def caption(self) -> str:
        return self._fields.get('CaptionTextValue_v1')

    @property
    def element(self) -> bytes:
        return self._fields.get('ElementValue_v1')

    def __str__(self):
        return f'<Focused ElementCaption: {self.caption}>'


class AXAuditElement_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)

    @property
    def identifier(self) -> bytes:
        return self._fields['PlatformElementValue_v1'].NSdata

    def __repr__(self):
        return f'<Element: {self.identifier}>'


class AXAuditInspectorSection_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)


class AXAuditElementAttribute_v1(SerializedObject):
    def __init__(self, fields):
        super().__init__(fields)


class AXAuditDeviceSetting_v1(SerializedObject):
    FIELDS = ('IdentiifierValue_v1', 'CurrentValueNumber_v1')

    def __init__(self, fields):
        super().__init__(fields)
        for k in self.FIELDS:
            if k not in self._fields:
                self._fields[k] = None

    @property
    def key(self) -> str:
        return self._fields['IdentiifierValue_v1']

    @property
    def value(self) -> typing.Any:
        return self._fields['CurrentValueNumber_v1']

    def __str__(self) -> str:
        return f'<AXAuditDeviceSetting_v1 {self.key} = {self.value}>'


SERIALIZABLE_OBJECTS = {
    'AXAuditDeviceSetting_v1': AXAuditDeviceSetting_v1,
    'AXAuditInspectorFocus_v1': AXAuditInspectorFocus_v1,
    'AXAuditElement_v1': AXAuditElement_v1,
    'AXAuditInspectorSection_v1': AXAuditInspectorSection_v1,
    'AXAuditElementAttribute_v1': AXAuditElementAttribute_v1,
}


@dataclass
class Event:
    name: str
    data: SerializedObject


class Direction(Enum):
    Previous = 3
    Next = 4
    First = 5
    Last = 6


def deserialize_object(d):
    if not isinstance(d, dict):
        if isinstance(d, list):
            return [deserialize_object(x) for x in d]
        return d

    if 'ObjectType' not in d:
        # simple dictionary
        new_dict = {}
        for k, v in d.items():
            new_dict[k] = deserialize_object(v)
        return new_dict

    if d['ObjectType'] == 'passthrough':
        return deserialize_object(d['Value'])
    else:
        return SERIALIZABLE_OBJECTS[d['ObjectType']](deserialize_object(d['Value']))


class AccessibilityAudit(RemoteServer):
    SERVICE_NAME = 'com.apple.accessibility.axAuditDaemon.remoteserver'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME, remove_ssl_context=True)

        # flush previously received messages
        self.recv_plist()
        if Version(lockdown.product_version) >= Version('15.0'):
            self.recv_plist()

    @property
    def capabilities(self) -> typing.List[str]:
        self.broadcast.deviceCapabilities()
        return self.recv_plist()[0]

    @property
    def settings(self) -> typing.List[AXAuditDeviceSetting_v1]:
        self.broadcast.deviceAccessibilitySettings()
        return deserialize_object(self.recv_plist()[0])

    def perform_handshake(self) -> None:
        # this service acts differently from others, requiring no handshake
        pass

    def set_app_monitoring_enabled(self, value: bool) -> None:
        self.broadcast.deviceSetAppMonitoringEnabled_(MessageAux().append_obj(value), expects_reply=False)

    def set_monitored_event_type(self, event_type: int = None) -> None:
        if event_type is None:
            event_type = 0
        self.broadcast.deviceInspectorSetMonitoredEventType_(MessageAux().append_obj(event_type), expects_reply=False)

    def set_show_ignored_elements(self, value: bool) -> None:
        self.broadcast.deviceInspectorShowIgnoredElements_(MessageAux().append_obj(int(value)), expects_reply=False)

    def set_show_visuals(self, value: bool) -> None:
        self.broadcast.deviceInspectorShowVisuals_(MessageAux().append_obj(int(value)), expects_reply=False)

    def iter_events(self, app_monitoring_enabled=True, monitored_event_type: int = None) -> \
            typing.Generator[Event, None, None]:

        self.set_app_monitoring_enabled(app_monitoring_enabled)
        self.set_monitored_event_type(monitored_event_type)

        while True:
            message = self.recv_plist()
            if message[1] is None:
                continue
            data = [x['value'] for x in message[1]]
            yield Event(name=message[0], data=deserialize_object(data))

    def move_focus_next(self) -> None:
        self.move_focus(Direction.Next)

    def perform_press(self, element: bytes) -> None:
        """ simulate click (can be used only for processes with task_for_pid-allow """
        element = {
            'ObjectType': 'AXAuditElement_v1',
            'Value': {
                'ObjectType': 'passthrough',
                'Value': {
                    'PlatformElementValue_v1': {
                        'ObjectType': 'passthrough'
                    },
                    'Value': element,
                }
            }
        }

        action = {
            'ObjectType': 'AXAuditElementAttribute_v1',
            'Value': {
                'ObjectType': 'passthrough',
                'Value': {
                    'AttributeNameValue_v1': {
                        'ObjectType': 'passthrough',
                        'Value': 'AXAction-2010',
                    },
                    'DisplayAsTree_v1': {
                        'ObjectType': 'passthrough',
                        'Value': 0,
                    },
                    'HumanReadableNameValue_v1': {
                        'ObjectType': 'passthrough',
                        'Value': 'Activate',
                    },
                    'IsInternal_v1': {
                        'ObjectType': 'passthrough',
                        'Value': 0,
                    },
                    'PerformsActionValue_v1': {
                        'ObjectType': 'passthrough',
                        'Value': 1,
                    },
                    'SettableValue_v1': {
                        'ObjectType': 'passthrough',
                        'Value': 0,
                    },
                    'ValueTypeValue_v1': {
                        'ObjectType': 'passthrough',
                        'Value': 1,
                    },
                }
            }
        }

        self.broadcast.deviceElement_performAction_withValue_(
            MessageAux().append_obj(element).append_obj(action).append_obj(0), expects_reply=False)

    def move_focus(self, direction: Direction) -> None:
        options = {
            'ObjectType': 'passthrough',
            'Value': {
                'allowNonAX': {
                    'ObjectType': 'passthrough',
                    'Value': 0,
                },
                'direction': {
                    'ObjectType': 'passthrough',
                    'Value': direction.value,
                },
                'includeContainers': {
                    'ObjectType': 'passthrough',
                    'Value': 1,
                }
            }
        }

        self.broadcast.deviceInspectorMoveWithOptions_(MessageAux().append_obj(options), expects_reply=False)

    def set_setting(self, name: str, value: typing.Any) -> None:
        setting = {'ObjectType': 'AXAuditDeviceSetting_v1',
                   'Value': {'ObjectType': 'passthrough',
                             'Value': {'CurrentValueNumber_v1': {'ObjectType': 'passthrough',
                                                                 'Value': True},
                                       'EnabledValue_v1': {'ObjectType': 'passthrough', 'Value': True},
                                       'IdentiifierValue_v1': {'ObjectType': 'passthrough',
                                                               'Value': name},
                                       'SettingTypeValue_v1': {'ObjectType': 'passthrough', 'Value': 3},
                                       'SliderTickMarksValue_v1': {'ObjectType': 'passthrough', 'Value': 0}}}}
        self.broadcast.deviceUpdateAccessibilitySetting_withValue_(
            MessageAux().append_obj(setting).append_obj({'ObjectType': 'passthrough', 'Value': value}),
            expects_reply=False)



import asyncio
import hashlib
import json
import logging
from datetime import datetime
from functools import partial

from pymobiledevice3.services.web_protocol.cdp_screencast import ScreenCast

logger = logging.getLogger(__name__)

NETWORK_RESOURCE_TYPES = ['Document', 'Stylesheet', 'Image', 'Media', 'Font', 'Script', 'TextTrack', 'XHR', 'Fetch',
                          'EventSource', 'WebSocket', 'Manifest', 'SignedExchange', 'Ping', 'CSPViolationReport',
                          'Preflight', 'Other']

LOG_MESSAGE_SOURCES = {
    'xml': 'xml',
    'javascript': 'javascript',
    'network': 'network',
    'console-api': 'javascript',
    'storage': 'storage',
    'appcache': 'appcache',
    'rendering': 'rendering',
    'css': 'rendering',
    'security': 'security',
    'deprecation': 'deprecation',
    'worker': 'worker',
    'violation': 'violation',
    'intervention': 'intervention',
    'recommendation': 'recommendation',
    'other': 'other',
    'content-blocker': 'other',
    'media': 'other',
    'mediasource': 'other',
    'webrtc': 'other',
    'itp-debug': 'other',
    'ad-click-attribution': 'other',
}

LOG_MESSAGE_LEVELS = {
    'log': 'info',
    'info': 'info',
    'warning': 'warning',
    'error': 'error',
    'debug': 'verbose',

}

DEBUGGER_PAUSED_REASON = {
    'XHR': 'XHR',
    'Fetch': 'other',
    'DOM': 'DOM',
    'AnimationFrame': 'other',
    'Interval': 'other',
    'Listener': 'EventListener',
    'Timeout': 'other',
    'exception': 'exception',
    'assert': 'assert',
    'CSPViolation': 'CSPViolation',
    'DebuggerStatement': 'debugCommand',
    'Breakpoint': 'instrumentation',
    'PauseOnNextStatement': 'instrumentation',
    'Microtask': 'other',
    'BlackboxedScript': 'other',
    'other': 'other',
}


class CdpTarget:
    def __init__(self, protocol, target_id: str):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        :param target_id: Target id.
        """
        self.protocol = protocol
        self.target_id = target_id
        self.frame = {}
        self.session_id = protocol.id_
        self.app_id = protocol.app.id_
        self.page_id = protocol.page.id_
        self.output_queue = asyncio.Queue()
        self.input_queue = asyncio.Queue()
        self.screencast = None
        self.from_cdp_special_messages_methods = {
            'Audits.enable': self._audits_enable,
            'DOM.getBoxModel': self._dom_get_box_model,
            'DOM.enable': partial(self._simple_response, value=None),
            'DOM.getNodeForLocation': self._dom_get_node_for_location,
            'DOM.getNodesForSubtreeByStyle': self._dom_get_nodes_for_subtree_by_style,
            'Log.clear': self._log_clear,
            'Log.disable': self._log_disable,
            'Log.enable': self._log_enable,
            'Log.startViolationsReport': partial(self._simple_response, value=None),
            'Page.getNavigationHistory': self._page_get_navigation_history,
            'Page.startScreencast': self._page_start_screencast,
            'Page.stopScreencast': self._page_stop_screencast,
            'Page.screencastFrameAck': self._page_screencast_frame_ack,
            'Page.getResourceTree': self._page_get_resource_tree,
            'Emulation.setEmulatedMedia': self._emulation_set_emulated_media,
            'Emulation.setTouchEmulationEnabled': partial(self._simple_response, value=None),
            'Emulation.setFocusEmulationEnabled': partial(self._simple_response, value=None),
            'Emulation.setEmulatedVisionDeficiency': partial(self._simple_response, value=None),
            'Emulation.setAutoDarkModeOverride': self._emulation_set_auto_dark_mode_override,
            'Emulation.setEmitTouchEventsForMouse': partial(self._simple_response, value=None),
            'Debugger.setAsyncCallStackDepth': partial(self._simple_response, value=True),
            'Debugger.setBlackboxPatterns': self._debugger_set_blackbox_patterns,
            'Debugger.setBreakpointByUrl': self._debugger_set_breakpoint_by_url,
            'DOMDebugger.setBreakOnCSPViolation': partial(self._simple_response, value=None),
            'DOMDebugger.getEventListeners': self._domdebugger_get_event_listeners,
            'Network.setCacheDisabled': self._network_set_cache_disabled,
            'Network.loadNetworkResource': self._network_load_network_resource,
            'Network.setAttachDebugStack': partial(self._simple_response, value=None),
            'Network.clearAcceptedEncodingsOverride': partial(self._simple_response, value=None),
            'ServiceWorker.enable': self._service_worker_enable,
            'HeapProfiler.enable': partial(self._simple_response, value=None),
            'Overlay.setShowGridOverlays': partial(self._simple_response, value=None),
            'Overlay.setShowFlexOverlays': partial(self._simple_response, value=None),
            'Overlay.setShowScrollSnapOverlays': partial(self._simple_response, value=None),
            'Overlay.setShowContainerQueryOverlays': partial(self._simple_response, value=None),
            'Overlay.setShowIsolatedElements': partial(self._simple_response, value=None),
            'Overlay.hideHighlight': partial(self._simple_response, value=None),
            'Overlay.highlightNode': self._overlay_highlight_node,
            'Overlay.enable': partial(self._simple_response, value=None),
            'Overlay.setShowViewportSizeOnResize': partial(self._simple_response, value=None),
            'Overlay.setPausedInDebuggerMessage': partial(self._simple_response, value=None),
            'Runtime.runIfWaitingForDebugger': partial(self._simple_response, value=None),
            'Runtime.compileScript': self._runtime_compile_script,
            'Runtime.getIsolateId': self._runtime_get_isolate_id,
            'Profiler.enable': partial(self._simple_response, value=None),
            'Target.setAutoAttach': self._target_set_auto_attach,
            'Target.setDiscoverTargets': partial(self._simple_response, value=None),
            'Target.setRemoteLocations': partial(self._simple_response, value=None),
            'CSS.trackComputedStyleUpdates': partial(self._simple_response, value=None),
            'CSS.takeComputedStyleUpdates': self._css_take_computed_style_updates,
            'CSS.addRule': self._css_add_rule,
            'Input.emulateTouchFromMouseEvent': self._input_emulate_touch_from_mouse_event,
            'Input.dispatchKeyEvent': self._input_dispatch_key_event,

        }
        self.to_cdp_special_messages_methods = {
            'Target.targetCreated': self._target_created,
            'Target.targetDestroyed': self._target_destroyed,
            'Target.dispatchMessageFromTarget': self._target_dispatch_message_from_target,
            'Target.didCommitProvisionalTarget': self._target_did_commit_provisional_target,
        }
        self.to_cdp_special_dispatched_messages_methods = {
            'Debugger.scriptParsed': self._debugger_script_parsed,
            'Debugger.scriptFailedToParse': self._debugger_script_failed_to_parse,
            'Debugger.paused': self._debugger_paused,
            'Debugger.globalObjectCleared': self._debugger_global_object_cleared,
            'Page.defaultAppearanceDidChange': self._page_default_appearance_did_change,
            'Runtime.executionContextCreated': self._runtime_execution_context_created,
            'Console.messageAdded': self._console_message_added,
            'Network.responseReceived': self._network_response_received,
            'Network.loadingFinished': self._network_loading_finished,
        }
        self._waiting_for_id = 0
        self._input_task = asyncio.create_task(self._input_loop())
        self._receiving_task = asyncio.create_task(self._receive_loop())
        self._script_source_to_context_id = {}
        self._default_execution_id = 0

    @classmethod
    async def create(cls, protocol):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        """
        await protocol.inspector.setup_inspector_socket(protocol.id_, protocol.app.id_, protocol.page.id_)
        while not protocol.inspector.wir_events:
            await asyncio.sleep(0)
        created = protocol.inspector.wir_events.pop(0)
        while 'targetInfo' not in created['params']:
            created = protocol.inspector.wir_events.pop(0)
        target_id = created['params']['targetInfo']['targetId']
        logger.info(f'Created: {target_id}')
        target = cls(protocol, target_id)
        await target.output_queue.put(created)
        return target

    async def send(self, message):
        """
        Send message from devtools to the target.
        """
        await self.input_queue.put(message)

    async def receive(self):
        """
        Get message from the target to the devtools.
        """
        return await self.output_queue.get()

    async def wait_for_event_id(self, id_: int):
        """
        Wait for a message with a specific id from the target.
        :param id_: Message id to wait for.
        """
        while True:
            for i in range(len(self.protocol.inspector.wir_events)):
                message = self.protocol.inspector.wir_events[i]
                if message['method'] != 'Target.dispatchMessageFromTarget':
                    continue
                message = json.loads(message['params']['message'])
                if message.get('id', '') != id_:
                    continue
                del self.protocol.inspector.wir_events[i]
                return message
            await asyncio.sleep(0)

    async def send_message_with_result(self, id_, method, params):
        """
        Send a message to the target and wait for response.
        """
        self._waiting_for_id += 1
        await self._send_message_to_target({'id': id_, 'method': method, 'params': params})
        result = await self.wait_for_event_id(id_)
        self._waiting_for_id -= 1
        return result

    async def evaluate_and_result(self, id_, expression):
        """
        Evaluate Javascript expression.
        """
        logger.debug('Evaluating: ', expression)
        params = {'expression': expression}
        data = await self.send_message_with_result(id_, 'Runtime.evaluate', params)
        logger.debug('Evaluated: ', data)
        result = data['result']['result']
        if result['type'] == 'string':
            return result['value']
        elif result['type'] == 'undefined':
            return None
        elif result['type'] == 'object':
            return result
        else:
            logger.debug('Unknown type: ', result)
            return result

    async def _input_loop(self):
        while True:
            message = await self.input_queue.get()
            if message['method'] in self.from_cdp_special_messages_methods:
                await self.from_cdp_special_messages_methods[message['method']](message)
            else:
                await self._send_message_to_target(message)

    async def _receive_loop(self):
        while True:
            if self._waiting_for_id or not self.protocol.inspector.wir_events:
                await asyncio.sleep(0)
                continue
            message = self.protocol.inspector.wir_events.pop(0)
            await self._to_output_queue(message)

    async def _to_output_queue(self, message):
        if message['method'] in self.to_cdp_special_messages_methods:
            await self.to_cdp_special_messages_methods[message['method']](message)
        else:
            logger.error('Error!!!!!!!!!!!!', message)
            raise RuntimeError()

    async def _send_message_to_target(self, message):
        await self.protocol.send_command('Target.sendMessageToTarget', targetId=self.target_id,
                                         message=json.dumps(message))

    async def _simple_response(self, message, value):
        await self.output_queue.put({'id': message['id'], 'result': {'result': value}})

    async def _audits_enable(self, message):
        message['method'] = 'Audit.setup'
        await self._send_message_to_target(message)

    async def _dom_get_box_model(self, message):
        message['method'] = 'DOM.highlightNode'
        message['params']['highlightConfig'] = {
            'showInfo': True,
            'contentColor': {'r': 111, 'g': 168, 'b': 220, 'a': 0.66},
            'paddingColor': {'r': 147, 'g': 196, 'b': 125, 'a': 0.55},
            'borderColor': {'r': 255, 'g': 229, 'b': 153, 'a': 0.66},
            'marginColor': {'r': 246, 'g': 178, 'b': 107, 'a': 0.66},
        }
        await self._send_message_to_target(message)

    async def object_id_to_node_id(self, object_id, id_):
        node = await self.send_message_with_result(id_, 'DOM.requestNode', {'objectId': object_id})
        return node['result']['nodeId']

    async def _dom_get_node_for_location(self, message):
        x, y = message['params']['x'], message['params']['y']
        obj = await self.evaluate_and_result(message['id'], f'document.elementFromPoint({x},{y})')
        if 'objectId' not in obj:
            await self._simple_response(message, None)
            return
        result = {'nodeId': await self.object_id_to_node_id(obj['objectId'], message['id'])}
        await self.output_queue.put({'id': message['id'], 'result': result})

    async def _dom_get_nodes_for_subtree_by_style(self, message):
        object_id = (await self.send_message_with_result(
            message['id'], 'DOM.resolveNode', {'nodeId': message['params']['nodeId']}
        ))['result']['object']['objectId']
        result = await self.send_message_with_result(message['id'], 'Runtime.callFunctionOn', {
            'objectId': object_id,
            'functionDeclaration': (
                'function(styles) {'
                '   const result = new Set();'
                '   var all = this.getElementsByTagName("*");'
                '   for (var elem_i=0; elem_i < all.length; elem_i++) {'
                '       for (var style_i in styles) {'
                '           if (window.getComputedStyle(all[elem_i]).getPropertyValue(styles[style_i].name) ==='
                ' styles[style_i].value) {'
                '               result.add(all[elem_i]);'
                '               break;'
                '           }'
                '       }'
                '   }'
                '   return result;'
                '}'
            ),
            'arguments': [{'value': message['params']['computedStyles']}]
        })
        result = await self.send_message_with_result(message['id'], 'Runtime.getCollectionEntries',
                                                     {'objectId': result['result']['result']['objectId']})
        nodes = await asyncio.gather(
            *[self.object_id_to_node_id(obj['value']['objectId'], message['id'] - i) for i, obj in
              enumerate(result['result']['entries'])], return_exceptions=True)
        nodes = [n for n in nodes if isinstance(n, int)]
        await self.output_queue.put({'id': message['id'], 'result': {'nodeIds': nodes}})

    async def _log_clear(self, message):
        message['method'] = 'Console.clearMessages'
        await self._send_message_to_target(message)

    async def _log_disable(self, message):
        message['method'] = 'Console.disable'
        await self._send_message_to_target(message)

    async def _log_enable(self, message):
        message['method'] = 'Console.enable'
        await self._send_message_to_target(message)

    async def _page_get_navigation_history(self, message):
        href = await self.evaluate_and_result(message['id'], 'window.location.href')
        title = await self.evaluate_and_result(message['id'], 'document.title')
        await self.output_queue.put({
            'id': message['id'],
            'result': {
                'currentIndex': 0,
                'entries': [{'id': 0, 'url': href, 'title': title}]
            }
        })

    async def _page_start_screencast(self, message):
        params = message['params']
        self.screencast = ScreenCast(self, params['format'], params['quality'], params['maxWidth'], params['maxHeight'])
        await self.screencast.start(message['id'])
        await self._simple_response(message, None)

    async def _page_stop_screencast(self, message):
        if self.screencast is not None:
            await self.screencast.stop()
            self.screencast = None
        await self._simple_response(message, None)

    async def _page_screencast_frame_ack(self, message):
        if self.screencast is not None:
            self.screencast.ack(message['params']['sessionId'])
        await self._simple_response(message, None)

    async def _page_get_resource_tree(self, message):
        result = await self.send_message_with_result(message['id'], message['method'], message['params'])
        self.frame = result['result']['frameTree']['frame']
        await self.output_queue.put(result)

    async def _emulation_set_emulated_media(self, message):
        message['method'] = 'Page.setEmulatedMedia'
        await self._send_message_to_target(message)

    async def _emulation_set_auto_dark_mode_override(self, message):
        message['method'] = 'Page.setForcedAppearance'
        params = message['params']
        if not params:
            await self._simple_response(message, None)
            return
        message['params'] = {'appearance': 'Dark' if params['enabled'] else 'Light'}
        await self._send_message_to_target(message)

    async def _debugger_set_blackbox_patterns(self, message):
        for pattern in message['params']['patterns']:
            await self.send_message_with_result(message['id'], 'Debugger.setShouldBlackboxURL',
                                                {'url': pattern, 'shouldBlackbox': True})
        await self._simple_response(message, None)

    async def _debugger_set_breakpoint_by_url(self, message):
        condition = message['params'].pop('condition', '')
        if condition:
            message['params']['options']['condition'] = condition
        await self._send_message_to_target(message)

    async def _domdebugger_get_event_listeners(self, message):
        node = {'nodeId': await self.object_id_to_node_id(message['params']['objectId'], message['id'])}
        listeners = await self.send_message_with_result(message['id'], 'DOM.getEventListenersForNode', node)
        if 'error' in listeners:
            await self._simple_response(message, None)
            return
        listeners_out = []
        for listener in listeners['result']['listeners']:
            data = {
                'type': listener['type'],
                'useCapture': listener['useCapture'],
                'passive': listener.get('passive', False),
                'once': listener.get('once', False),
            }
            if 'location' in listener:
                data['scriptId'] = listener['location']['scriptId']
                data['lineNumber'] = listener['location']['lineNumber']
                data['columnNumber'] = listener['location']['columnNumber']
            listeners_out.append(data)
        await self.output_queue.put({'id': message['id'], 'result': {'listeners': listeners_out}})

    async def _network_set_cache_disabled(self, message):
        message['method'] = 'Network.setResourceCachingDisabled'
        message['params'] = {'disabled': message['params']['cacheDisabled']}
        await self._send_message_to_target(message)

    async def _network_load_network_resource(self, message):
        await self.output_queue.put({'id': message['id'], 'result': {'resource': {'success': True}}})

    async def _service_worker_enable(self, message):
        message['method'] = 'Worker.enable'
        await self._send_message_to_target(message)

    async def _overlay_highlight_node(self, message):
        message['method'] = 'DOM.highlightNode'
        await self._send_message_to_target(message)

    async def _runtime_compile_script(self, message):
        self._script_source_to_context_id[message['params']['expression']] = message['params']['executionContextId']
        response = await self.send_message_with_result(message['id'], 'Runtime.parse',
                                                       {'source': message['params']['expression']})
        if response['result']['result'] == 'none':
            await self._simple_response(message, None)
            return
        lines = message['params']['expression'][:response['result']['range']['endOffset']].splitlines()
        lines = lines if lines else ['']
        await self.output_queue.put({'id': message['id'], 'result': {
            'exceptionDetails': {
                'exceptionId': 1,
                'text': response['result']['message'],
                'lineNumber': len(lines) - 1,
                'columnNumber': len(lines[-1]) - 1
            }
        }})

    async def _runtime_get_isolate_id(self, message):
        await self.output_queue.put({'id': message['id'], 'result': {'id': self._default_execution_id}})

    async def _target_set_auto_attach(self, message):
        await self._simple_response(message, None)
        await self.output_queue.put({'method': 'Target.attachedToTarget', 'params': {
            'sessionId': self.protocol.id_, 'targetInfo': {'targetId': self.target_id}, 'waitingForDebugger': True
        }})

    async def _css_take_computed_style_updates(self, message):
        await self.output_queue.put({
            'id': message['id'],
            'result': {'nodeIds': []}
        })

    async def _css_add_rule(self, message):
        message['params']['selector'] = message['params']['ruleText'].split('{')[0]
        await self._send_message_to_target(message)

    async def _input_emulate_touch_from_mouse_event(self, message):
        params = message['params']
        if params['type'] == 'mouseWheel':
            delta_x, delta_y = params['deltaX'] // self.screencast.scale, params['deltaY'] // self.screencast.scale
            await self.evaluate_and_result(message['id'], f'window.scrollBy({-delta_x}, {-delta_y})')
        elif params['type'] == 'mouseReleased':
            pass
        else:
            modifiers = params['modifiers']
            x, y = params['x'] // self.screencast.scale, params['y'] // self.screencast.scale
            event_params = {
                'screenX': x, 'screenY': y, 'clientX': 0, 'clientY': 0, 'altKey': bool(modifiers & 1),
                'ctrlKey': bool(modifiers & 2), 'metaKey': bool(modifiers & 4), 'shiftKey': bool(modifiers & 8),
                'button': params['button'], 'bubbles': True, 'cancelable': False,
            }
            type_ = {'mousePressed': 'click', 'mouseReleased': 'click', 'mouseMoved': 'mousemove'}[params['type']]
            event_params = json.dumps(event_params)
            simulate_mouse_event = (
                'function simulateMouseEvent(type){'
                f'const element = document.elementFromPoint({x}, {y});'
                f'const e = new MouseEvent(type, JSON.parse(\'{event_params}\'));'
                'element.dispatchEvent(e);'
                'element.focus();'
                'return e;}'
            )
            await self.evaluate_and_result(message['id'], f'({simulate_mouse_event})("{type_}")')
            if type_ == 'click':
                await self.evaluate_and_result(message['id'], f'({simulate_mouse_event})("mouseup")')

        await self._simple_response(message, None)

    async def _input_dispatch_key_event(self, message):
        params = message['params']
        key = params['key']
        if params['type'] == 'keyUp' and key == 'Backspace':
            manipulation = 'document.activeElement.value = document.activeElement.value.slice(0, -1);'
        elif params['type'] == 'char' and key == 'Enter':
            manipulation = (
                'var tagName = document.activeElement.tagName.toLowerCase();'
                'if (tagName === "textarea" || document.activeElement.isContentEditable) {'
                '    document.activeElement.value = document.activeElement.value + "\\n";'
                '} else {'
                '    const result = document.evaluate("./ancestor-or-self::form", document.activeElement, '
                '                                     null, XPathResult.FIRST_ORDERED_NODE_TYPE, null);'
                '    const e = result.singleNodeValue.ownerDocument.createEvent(\'Event\');'
                '    e.initEvent("submit", true, true);'
                '    if (result.singleNodeValue.dispatchEvent(e)) { result.singleNodeValue.submit() }'
                '}'
            )
        elif params['type'] == 'char':
            text = params['text']
            manipulation = f'document.activeElement.value = document.activeElement.value + "{text}";'
        else:
            await self._simple_response(message, None)
            return

        simulate_key_event = (
            'function isEditable(element) {'
            '    if (element.disabled || element.readOnly)'
            '        return false;'
            '    var tagName = element.tagName.toLowerCase();'
            '    if (tagName === "textarea" || element.isContentEditable)'
            '        return true;'
            '    if (tagName != "input")'
            '        return false;'
            '    switch (element.type) {'
            '    case "color": case "date": case "datetime-local": case "email": case "file": case "month": '
            '    case "number": case "password": case "range": case "search": case "tel": case "text": case "time": '
            '    case "url": case "week":'
            '        return true;'
            '    }'
            '    return false;'
            '}'
            'if (isEditable(document.activeElement)) {'
            f'{manipulation}'
            '}'
        )
        await self.evaluate_and_result(message['id'], simulate_key_event)
        await self._simple_response(message, None)

    async def _target_created(self, message):
        self.target_id = message['params']['targetInfo']['targetId']
        message['method'] = 'Target.targetInfoChanged'
        message['params']['targetInfo']['url'] = await self.evaluate_and_result(1, 'window.location.href')
        message['params']['targetInfo']['title'] = await self.evaluate_and_result(1, 'document.title')
        message['params']['targetInfo']['attached'] = message['params']['targetInfo'].pop('isProvisional')
        await self.output_queue.put(message)

    async def _target_destroyed(self, message):
        result = await self.send_message_with_result(1, 'Page.getResourceTree', {})
        self.frame = result['result']['frameTree']['frame']
        await self.output_queue.put({
            'method': 'Page.frameNavigated',
            'params': {
                'frame': self.frame,
            }
        })
        await self.output_queue.put({
            'method': 'Page.loadEventFired',
            'params': {
                'timestamp': datetime.now().strftime('%Y-%m-%dT%H:%M:%S.%f'),
            }
        })
        await self.output_queue.put({'method': 'DOM.documentUpdated'})

    async def _target_dispatch_message_from_target(self, message):
        message = json.loads(message['params']['message'])
        if 'error' in message:
            logger.error(message)
        if message.get('method', '') in self.to_cdp_special_dispatched_messages_methods:
            await self.to_cdp_special_dispatched_messages_methods[message['method']](message)
        else:
            if 'method' in message:
                logger.debug('DISPACHING', message['method'])
            await self.output_queue.put(message)

    async def _target_did_commit_provisional_target(self, message):
        pass

    async def _debugger_script_parsed(self, message):
        if not self._waiting_for_id:
            await self.output_queue.put(message)

    async def _debugger_script_failed_to_parse(self, message):
        message['params'] = {
            'endColumn': 0,
            'endLine': message['params']['errorLine'],
            'executionContextId': self._script_source_to_context_id[message['params']['scriptSource']],
            'startColumn': 0,
            'startLine': message['params']['startLine'],
            'url': message['params']['url'],
            'scriptId': self._script_source_to_context_id[message['params']['scriptSource']],
            'hash': hashlib.sha1(message['params']['scriptSource']).hexdigest(),
        }
        await self.output_queue.put(message)

    async def _debugger_paused(self, message):
        message['params']['reason'] = DEBUGGER_PAUSED_REASON[message['params']['reason']]
        if 'breakpointId' in message['params'].get('data', {}):
            message['params']['hitBreakpoints'] = [message['params']['data']['breakpointId']]
        await self.output_queue.put(message)

    async def _debugger_global_object_cleared(self, message):
        await self.output_queue.put({'method': 'DOM.documentUpdated'})

    async def _page_default_appearance_did_change(self, message):
        pass

    async def _runtime_execution_context_created(self, message):
        if message['params']['context']['type'] == 'normal':
            self._default_execution_id = message['params']['context']['id']
        message['params'] = {
            'context': {
                'id': message['params']['context']['id'],
                'origin': 'default',
                'name': '',
                'uniqueId': message['params']['context']['frameId'],
            }
        }
        await self.output_queue.put(message)

    async def _console_message_added(self, message):
        log_record = {
            'source': LOG_MESSAGE_SOURCES[message['params']['message']['source']],
            'level': LOG_MESSAGE_LEVELS[message['params']['message']['level']],
            'text': message['params']['message']['text'],
            'timestamp': datetime.now().timestamp(),
        }
        if 'url' in message['params']['message']:
            log_record['url'] = message['params']['message']['url']
        if 'line' in message['params']['message']:
            log_record['lineNumber'] = message['params']['message']['line']
        if 'networkRequestId' in message['params']['message']:
            log_record['networkRequestId'] = message['params']['message']['networkRequestId']

        await self.output_queue.put({'method': 'Log.entryAdded', 'params': {'entry': log_record}})

    async def _network_response_received(self, message):
        params = message['params']
        message['params'] = {
            'loaderId': params['loaderId'],
            'requestId': params['requestId'],
            'timestamp': params['timestamp'],
            'type': params['type'] if params['type'] in NETWORK_RESOURCE_TYPES else 'Other',
            'response': {
                'url': params['response']['url'],
                'status': params['response']['status'],
                'statusText': params['response']['statusText'],
                'headers': params['response']['headers'],
                'mimeType': params['response']['mimeType'],
                'connectionReused': False,
                'encodedDataLength': 0,
                'securityState': 'unknown',
            },
        }
        if 'frameId' in params:
            message['params']['frameId'] = params['frameId']
        await self.output_queue.put(message)

    async def _network_loading_finished(self, message):
        params = message['params']
        header_size = params['metrics'].get('responseHeaderBytesReceived', 0)
        body_size = params['metrics'].get('responseBodyBytesReceived', 0)
        message['params'] = {
            'encodedDataLength': header_size + body_size,
            'requestId': params['requestId'],
            'timestamp': params['timestamp'],
        }
        await self.output_queue.put(message)



import logging

import click
from tqdm import tqdm

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.mobilebackup2 import Mobilebackup2Service

source_option = click.option('--source', default='', help='The UDID of the source device.')
password_option = click.option('-p', '--password', default='', help='Backup password.')
backup_directory_arg = click.argument('backup-directory', type=click.Path(exists=True, file_okay=False))
backup_directory_option = click.option('-b', '--backup-directory', type=click.Path(exists=True, file_okay=False),
                                       default='.')

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ backup cli """
    pass


@cli.group()
def backup2():
    """ backup utils """
    pass


@backup2.command(cls=Command)
@click.argument('backup-directory', type=click.Path(file_okay=False))
@click.option('--full', is_flag=True, help=('Whether to do a full backup.'
                                            ' If full is True, any previous backup attempts will be discarded.'))
def backup(lockdown: LockdownClient, backup_directory, full):
    """
    Backup device.

    All backup data will be written to BACKUP_DIRECTORY, under a directory named with the device's udid.
    """
    backup_client = Mobilebackup2Service(lockdown)
    with tqdm(total=100, dynamic_ncols=True) as pbar:
        def update_bar(percentage):
            pbar.n = percentage
            pbar.refresh()

        backup_client.backup(full=full, backup_directory=backup_directory, progress_callback=update_bar)


@backup2.command(cls=Command)
@backup_directory_arg
@click.option('--system/--no-system', default=False, help='Restore system files.')
@click.option('--reboot/--no-reboot', default=True, help='Reboot the device when done.')
@click.option('--copy/--no-copy', default=True, help='Create a copy of backup folder before restoring.')
@click.option('--settings/--no-settings', default=True, help='Restore device settings.')
@click.option('--remove/--no-remove', default=False, help='Remove items which aren\'t being restored.')
@password_option
@source_option
def restore(lockdown: LockdownClient, backup_directory, system, reboot, copy, settings, remove, password, source):
    """
    Restore a backup to a device.

    The backup will be restored from a directory with the device udid under BACKUP_DIRECTORY.
    """
    backup_client = Mobilebackup2Service(lockdown)
    with tqdm(total=100, dynamic_ncols=True) as pbar:
        def update_bar(percentage):
            pbar.n = percentage
            pbar.refresh()

        backup_client.restore(backup_directory=backup_directory, progress_callback=update_bar, system=system,
                              reboot=reboot, copy=copy, settings=settings, remove=remove, password=password,
                              source=source)


@backup2.command(cls=Command)
@backup_directory_arg
@source_option
def info(lockdown: LockdownClient, backup_directory, source):
    """
    Print information about a backup.
    """
    backup_client = Mobilebackup2Service(lockdown)
    print(backup_client.info(backup_directory=backup_directory, source=source))


@backup2.command('list', cls=Command)
@backup_directory_arg
@source_option
def list_(lockdown: LockdownClient, backup_directory, source):
    """
    List all file in the backup in a CSV format.
    """
    backup_client = Mobilebackup2Service(lockdown)
    print(backup_client.list(backup_directory=backup_directory, source=source))


@backup2.command(cls=Command)
@backup_directory_arg
@password_option
@source_option
def unback(lockdown: LockdownClient, backup_directory, password, source):
    """
    Convert all files in the backup to the correct directory hierarchy.
    """
    backup_client = Mobilebackup2Service(lockdown)
    backup_client.unback(backup_directory=backup_directory, password=password, source=source)


@backup2.command(cls=Command)
@click.argument('domain-name')
@click.argument('relative-path')
@backup_directory_arg
@password_option
@source_option
def extract(lockdown: LockdownClient, domain_name, relative_path, backup_directory, password, source):
    """
    Extract a file from the backup.

    The file that belongs to the domain DOMAIN_NAME and located on the device in the path RELATIVE_PATH,
    will be extracted to the BACKUP_DIRECTORY.
    """
    backup_client = Mobilebackup2Service(lockdown)
    backup_client.extract(domain_name, relative_path, backup_directory=backup_directory, password=password,
                          source=source)


@backup2.command(cls=Command)
@click.argument('mode', type=click.Choice(['on', 'off'], case_sensitive=False))
@click.argument('password')
@backup_directory_option
def encryption(lockdown: LockdownClient, backup_directory, mode, password):
    """
    Set backup encryption on / off.

    When on, PASSWORD will be the new backup password.
    When off, PASSWORD is the current backup password.
    """
    backup_client = Mobilebackup2Service(lockdown)
    should_encrypt = mode.lower() == 'on'
    if should_encrypt == backup_client.will_encrypt:
        logger.error('Encryption already ' + ('on!' if should_encrypt else 'off!'))
        return
    if should_encrypt:
        backup_client.change_password(backup_directory, new=password)
    else:
        backup_client.change_password(backup_directory, old=password)


@backup2.command(cls=Command)
@click.argument('old-password')
@click.argument('new-password')
@backup_directory_option
def change_password(lockdown: LockdownClient, old_password, new_password, backup_directory):
    """
    Change the backup password.
    """
    backup_client = Mobilebackup2Service(lockdown)
    if not backup_client.will_encrypt:
        logger.error('Encryption is not turned on!')
        return
    backup_client.change_password(backup_directory, old=old_password, new=new_password)


@backup2.command(cls=Command)
@backup_directory_arg
def erase_device(lockdown: LockdownClient, backup_directory):
    """
    Erase all data on the device.
    """
    backup_client = Mobilebackup2Service(lockdown)
    backup_client.erase_device(backup_directory)



import logging
import plistlib

import click

from pymobiledevice3.cli.cli_common import Command, CommandWithoutAutopair, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.heartbeat import HeartbeatService

logger = logging.getLogger(__name__)


@click.group()
def cli():
    """ apps cli """
    pass


@cli.group('lockdown')
def lockdown_group():
    """ lockdown options """
    pass


@lockdown_group.command('recovery', cls=Command)
def lockdown_recovery(lockdown: LockdownClient):
    """ enter recovery """
    print_json(lockdown.enter_recovery())


@lockdown_group.command('service', cls=Command)
@click.argument('service_name')
def lockdown_service(lockdown: LockdownClient, service_name):
    """ send-receive raw service messages """
    lockdown.start_service(service_name).shell()


@lockdown_group.command('info', cls=Command)
@click.option('-a', '--all', is_flag=True, help='include all domain information')
@click.option('--color/--no-color', default=True)
def lockdown_info(lockdown: LockdownClient, all, color):
    """ query all lockdown values """
    print_json(lockdown.all_domains if all else lockdown.all_values, colored=color)


@lockdown_group.command('get', cls=Command)
@click.argument('domain', required=False)
@click.argument('key', required=False)
@click.option('--color/--no-color', default=True)
def lockdown_get(lockdown: LockdownClient, domain, key, color):
    """ query lockdown values by their domain and key names """
    print_json(lockdown.get_value(domain=domain, key=key), colored=color)


@lockdown_group.command('set', cls=Command)
@click.argument('value')
@click.argument('domain', required=False)
@click.argument('key', required=False)
@click.option('--color/--no-color', default=True)
def lockdown_set(lockdown: LockdownClient, value, domain, key, color):
    """ set a lockdown value using python's eval() """
    print_json(lockdown.set_value(value=eval(value), domain=domain, key=key), colored=color)


@lockdown_group.command('remove', cls=Command)
@click.argument('domain')
@click.argument('key')
@click.option('--color/--no-color', default=True)
def lockdown_remove(lockdown: LockdownClient, domain, key, color):
    """ remove a domain/key pair """
    print_json(lockdown.remove_value(domain=domain, key=key), colored=color)


@lockdown_group.command('unpair', cls=CommandWithoutAutopair)
def lockdown_unpair(lockdown: LockdownClient):
    """ unpair from connected device """
    lockdown.unpair()


@lockdown_group.command('pair', cls=CommandWithoutAutopair)
def lockdown_pair(lockdown: LockdownClient):
    """ pair device """
    lockdown.pair()


@lockdown_group.command('save-pair-record', cls=CommandWithoutAutopair)
@click.argument('output', type=click.File('wb'))
def lockdown_save_pair_record(lockdown: LockdownClient, output):
    """ save pair record to specified location """
    if lockdown.pair_record is None:
        logger.error('no pairing record was found')
        return
    plistlib.dump(lockdown.pair_record, output)


@lockdown_group.command('date', cls=Command)
def lockdown_date(lockdown: LockdownClient):
    """ get device date """
    print(lockdown.date)


@lockdown_group.command('heartbeat', cls=Command)
def lockdown_heartbeat(lockdown: LockdownClient):
    """ start heartbeat service """
    HeartbeatService(lockdown).start()


@lockdown_group.command('language', cls=Command)
def lockdown_language(lockdown: LockdownClient):
    """ get current language settings """
    print(f'{lockdown.language} {lockdown.locale}')


@lockdown_group.command('device-name', cls=Command)
@click.argument('new_name', required=False)
def lockdown_device_name(lockdown: LockdownClient, new_name):
    """ get/set current device name """
    if new_name:
        lockdown.set_value(new_name, key='DeviceName')
    else:
        print(f'{lockdown.get_value(key="DeviceName")}')


@lockdown_group.command('wifi-connections', cls=Command)
@click.argument('state', type=click.Choice(['on', 'off']), required=False)
def lockdown_wifi_connections(lockdown: LockdownClient, state):
    """ get/set wifi connections state """
    if not state:
        # show current state
        print_json(lockdown.get_value(domain='com.apple.mobile.wireless_lockdown'))
    else:
        # enable/disable
        lockdown.enable_wifi_connections = state == 'on'



import plistlib
from enum import Enum
from typing import Mapping

from pymobiledevice3.exceptions import ProfileError
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.base_service import BaseService


class Purpose(Enum):
    PostSetupInstallation = 'PostSetupInstallation'


class MobileConfigService(BaseService):
    SERVICE_NAME = 'com.apple.mobile.MCInstall'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    def hello(self) -> None:
        self._send_recv({'RequestType': 'HelloHostIdentifier'})

    def flush(self) -> None:
        self._send_recv({'RequestType': 'Flush'})

    def get_stored_profile(self, purpose: Purpose = Purpose.PostSetupInstallation) -> Mapping:
        return self._send_recv({'RequestType': 'GetStoredProfile', 'Purpose': purpose.value})

    def store_profile(self, profile_data: bytes, purpose: Purpose = Purpose.PostSetupInstallation) -> None:
        self._send_recv({'RequestType': 'StoreProfile', 'ProfileData': profile_data, 'Purpose': purpose.value})

    def get_cloud_configuration(self) -> Mapping:
        return self._send_recv({'RequestType': 'GetCloudConfiguration'}).get('CloudConfiguration')

    def set_cloud_configuration(self, cloud_configuration: Mapping) -> None:
        self._send_recv({'RequestType': 'SetCloudConfiguration', 'CloudConfiguration': cloud_configuration})

    def establish_provisional_enrollment(self, nonce: bytes) -> None:
        self._send_recv({'RequestType': 'EstablishProvisionalEnrollment', 'Nonce': nonce})

    def set_wifi_power_state(self, state: bool) -> None:
        self._send_recv({'RequestType': 'SetWiFiPowerState', 'PowerState': state})

    def erase_device(self, preserve_data_plan: bool, disallow_proximity_setup: bool) -> None:
        try:
            self._send_recv({'RequestType': 'EraseDevice', 'PreserveDataPlan': preserve_data_plan,
                             'DisallowProximitySetup': disallow_proximity_setup})
        except ConnectionAbortedError:
            pass

    def get_profile_list(self) -> Mapping:
        return self._send_recv({'RequestType': 'GetProfileList'})

    def install_profile(self, payload: bytes) -> None:
        self._send_recv({'RequestType': 'InstallProfile', 'Payload': payload})

    def remove_profile(self, identifier: str) -> None:
        profiles = self.get_profile_list()
        if not profiles:
            return
        if identifier not in profiles['ProfileMetadata']:
            self.logger.info(f'Trying to remove not installed profile: {identifier}')
            return
        meta = profiles['ProfileMetadata'][identifier]
        data = plistlib.dumps({'PayloadType': 'Configuration',
                               'PayloadIdentifier': identifier,
                               'PayloadUUID': meta['PayloadUUID'],
                               'PayloadVersion': meta['PayloadVersion']
                               })
        self._send_recv({'RequestType': 'RemoveProfile', 'ProfileIdentifier': data})

    def _send_recv(self, request: Mapping) -> Mapping:
        response = self.service.send_recv_plist(request)
        if response.get('Status', None) != 'Acknowledged':
            raise ProfileError(f'invalid response {response}')
        return response



import asyncio
import json
import logging
from typing import Mapping

from pymobiledevice3.exceptions import InspectorEvaluateError
from pymobiledevice3.services.web_protocol.session_protocol import SessionProtocol

logger = logging.getLogger(__name__)


class InspectorSession:

    def __init__(self, protocol: SessionProtocol, target_id: str):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        """
        self.protocol = protocol
        self.target_id = target_id
        self.message_id = 1
        self._dispatch_message_responses = {}

        self.response_methods = {
            'Target.targetCreated': self._target_created,
            'Target.targetDestroyed': self._target_destroyed,
            'Target.dispatchMessageFromTarget': self._target_dispatch_message_from_target,
            'Target.didCommitProvisionalTarget': self._target_did_commit_provisional_target,
        }

        self._receive_task = asyncio.create_task(self._receive_loop())

    @classmethod
    async def create(cls, protocol: SessionProtocol):
        """
        :param pymobiledevice3.services.web_protocol.session_protocol.SessionProtocol protocol: Session protocol.
        """
        await protocol.inspector.setup_inspector_socket(protocol.id_, protocol.app.id_, protocol.page.id_)
        while not protocol.inspector.wir_events:
            await asyncio.sleep(0)
        created = protocol.inspector.wir_events.pop(0)
        while 'targetInfo' not in created['params']:
            created = protocol.inspector.wir_events.pop(0)
        target_id = created['params']['targetInfo']['targetId']
        logger.info(f'Created: {target_id}')
        target = cls(protocol, target_id)
        return target

    def set_target_id(self, target_id):
        self.target_id = target_id
        logger.info(f'Changed to: {target_id}')

    async def runtime_enable(self):
        await self.send_and_receive({'method': 'Runtime.enable', 'params': {}})

    async def runtime_evaluate(self, exp: str):
        # if the expression is dict, it's needed to be in ()
        exp = exp.strip()
        if exp:
            if exp[0] == '{' and exp[-1] == '}':
                exp = f'({exp})'

        response = await self.send_and_receive({'method': 'Runtime.evaluate',
                                                'params': {
                                                    'expression': exp,
                                                    'objectGroup': 'console',
                                                    'includeCommandLineAPI': True,
                                                    'silent': False,
                                                    'returnByValue': True,
                                                    'generatePreview': False,
                                                    'userGesture': True,
                                                    'awaitPromise': False,
                                                    'replMode': True,
                                                    'allowUnsafeEvalBlockedByCSP': False,
                                                    'uniqueContextId': '0.1'}
                                                })

        return self._parse_runtime_evaluate(response)

    async def navigate_to_url(self, url: str):
        return await self.runtime_evaluate(exp=f'window.location = "{url}"')

    async def send_and_receive(self, message: Mapping) -> Mapping:
        message_id = await self.send_message_to_target(message)
        return await self.receive_response_by_id(message_id)

    async def send_message_to_target(self, message: Mapping) -> int:
        message['id'] = self.message_id
        self.message_id += 1
        await self.protocol.send_command('Target.sendMessageToTarget', targetId=self.target_id,
                                         message=json.dumps(message))
        return message['id']

    async def _receive_loop(self):
        while True:
            while not self.protocol.inspector.wir_events:
                await asyncio.sleep(0)

            response = self.protocol.inspector.wir_events.pop(0)
            response_method = response['method']
            if response_method in self.response_methods:
                self.response_methods[response_method](response)
            else:
                logger.error('Unknown response method')

    async def receive_response_by_id(self, message_id: int) -> Mapping:
        while True:
            if message_id in self._dispatch_message_responses:
                return self._dispatch_message_responses.pop(message_id)
            await asyncio.sleep(0)

    @staticmethod
    def _parse_runtime_evaluate(response: Mapping):
        message = json.loads(response['params']['message'])
        if 'error' in message:
            details = message['error']['message']
            logger.error(details)
            raise InspectorEvaluateError(details)

        result = message['result']['result']
        if result.get('subtype', '') == 'error':
            details = result['description']
            logger.error(details)
            raise InspectorEvaluateError(details)
        elif result['type'] == 'undefined':
            pass
        else:
            return result['value']

    # response methods
    def _target_dispatch_message_from_target(self, response: Mapping):
        receive_message_id = json.loads(response['params']['message'])['id']
        self._dispatch_message_responses[receive_message_id] = response

    def _target_created(self, response: Mapping):
        pass

    def _target_destroyed(self, response: Mapping):
        pass

    def _target_did_commit_provisional_target(self, response: Mapping):
        self.set_target_id(response['params']['newTargetId'])



import click
import IPython

from pymobiledevice3.cli.cli_common import Command, print_json
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.springboard import SpringBoardServicesService

SHELL_USAGE = '''
Use `service` to access the service features
'''


@click.group()
def cli():
    """ apps cli """
    pass


@cli.group()
def springboard():
    """ springboard options """
    pass


@springboard.group()
def state():
    """ icons state options """
    pass


@state.command('get', cls=Command)
@click.option('--color/--no-color', default=True)
def state_get(lockdown: LockdownClient, color):
    """ get icon state """
    print_json(SpringBoardServicesService(lockdown=lockdown).get_icon_state(), colored=color)


@springboard.command('shell', cls=Command)
def springboard_shell(lockdown: LockdownClient):
    """ open a shell to communicate with SpringBoardServicesService """
    service = SpringBoardServicesService(lockdown=lockdown)
    IPython.embed(
        header=SHELL_USAGE,
        user_ns={
            'service': service,
        })


@springboard.command('icon', cls=Command)
@click.argument('bundle_id')
@click.argument('out', type=click.File('wb'))
def springboard_icon(lockdown: LockdownClient, bundle_id, out):
    """ get application's icon """
    out.write(SpringBoardServicesService(lockdown=lockdown).get_icon_pngdata(bundle_id))


@springboard.command('orientation', cls=Command)
def springboard_orientation(lockdown: LockdownClient):
    """ get screen orientation """
    print(SpringBoardServicesService(lockdown=lockdown).get_interface_orientation())


@springboard.command('wallpaper', cls=Command)
@click.argument('out', type=click.File('wb'))
def springboard_wallpaper(lockdown: LockdownClient, out):
    """ get wallpapaer """
    out.write(SpringBoardServicesService(lockdown=lockdown).get_wallpaper_pngdata())



#!/usr/bin/env python3
import plistlib
import time
import uuid
from contextlib import contextmanager, suppress
from datetime import datetime
from pathlib import Path

from pymobiledevice3.exceptions import AfcException, AfcFileNotFoundError, ConnectionTerminatedError, LockdownError, \
    PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient
from pymobiledevice3.services.afc import AFC_LOCK_EX, AFC_LOCK_UN, AfcService, afc_error_t
from pymobiledevice3.services.base_service import BaseService
from pymobiledevice3.services.device_link import DeviceLink
from pymobiledevice3.services.installation_proxy import InstallationProxyService
from pymobiledevice3.services.notification_proxy import NotificationProxyService
from pymobiledevice3.services.springboard import SpringBoardServicesService

SUPPORTED_VERSIONS = [2.0, 2.1]
ITUNES_FILES = [
    'ApertureAlbumPrefs', 'IC-Info.sidb', 'IC-Info.sidv', 'PhotosFolderAlbums', 'PhotosFolderName',
    'PhotosFolderPrefs', 'VoiceMemos.plist', 'iPhotoAlbumPrefs', 'iTunesApplicationIDs', 'iTunesPrefs',
    'iTunesPrefs.plist'
]
NP_SYNC_WILL_START = 'com.apple.itunes-mobdev.syncWillStart'
NP_SYNC_DID_START = 'com.apple.itunes-mobdev.syncDidStart'
NP_SYNC_LOCK_REQUEST = 'com.apple.itunes-mobdev.syncLockRequest'
NP_SYNC_DID_FINISH = 'com.apple.itunes-mobdev.syncDidFinish'


class Mobilebackup2Service(BaseService):
    SERVICE_NAME = 'com.apple.mobilebackup2'

    def __init__(self, lockdown: LockdownClient):
        super().__init__(lockdown, self.SERVICE_NAME)

    @property
    def will_encrypt(self):
        try:
            return self.lockdown.get_value('com.apple.mobile.backup', 'WillEncrypt')
        except LockdownError:
            return False

    def backup(self, full: bool = True, backup_directory='.', progress_callback=lambda x: None) -> None:
        """
        Backup a device.
        :param full: Whether to do a full backup. If full is True, any previous backup attempts will be discarded.
        :param backup_directory: Directory to write backup to.
        :param progress_callback: Function to be called as the backup progresses.
        The function shall receive the percentage as a parameter.
        """
        backup_directory = Path(backup_directory)
        device_directory = backup_directory / self.lockdown.udid
        device_directory.mkdir(exist_ok=True, mode=0o755, parents=True)

        with self.device_link(backup_directory) as dl, \
                NotificationProxyService(self.lockdown) as notification_proxy, \
                AfcService(self.lockdown) as afc:
            with self._backup_lock(afc, notification_proxy):
                # Initialize Info.plist
                info_plist = self.init_mobile_backup_factory_info(afc)
                with open(device_directory / 'Info.plist', 'wb') as fd:
                    plistlib.dump(info_plist, fd)

                # Initialize Status.plist file if doesn't exist.
                status_path = device_directory / 'Status.plist'
                if full or not status_path.exists():
                    with open(device_directory / 'Status.plist', 'wb') as fd:
                        plistlib.dump({
                            'BackupState': 'new',
                            'Date': datetime.utcnow(),
                            'IsFullBackup': full,
                            'Version': '3.3',
                            'SnapshotState': 'finished',
                            'UUID': str(uuid.uuid4()).upper(),
                        }, fd, fmt=plistlib.FMT_BINARY)

                # Create Manifest.plist if doesn't exist.
                manifest_path = device_directory / 'Manifest.plist'
                if full:
                    manifest_path.unlink(missing_ok=True)
                (device_directory / 'Manifest.plist').touch()

                dl.send_process_message({'MessageName': 'Backup', 'TargetIdentifier': self.lockdown.udid})
                dl.dl_loop(progress_callback)

    def restore(self, backup_directory='.', system: bool = False, reboot: bool = True, copy: bool = True,
                settings: bool = True, remove: bool = False, password: str = '', source: str = '',
                progress_callback=lambda x: None):
        """
        Restore a previous backup to the device.
        :param backup_directory: Path of the backup directory.
        :param system: Whether to restore system files.
        :param reboot: Reboot the device when done.
        :param copy: Create a copy of backup folder before restoring.
        :param settings: Restore device settings.
        :param remove: Remove items which aren't being restored.
        :param password: Password of the backup if it is encrypted.
        :param source: Identifier of device to restore its backup.
        :param progress_callback: Function to be called as the backup progresses.
        The function shall receive the current percentage of the progress as a parameter.
        """
        backup_directory = Path(backup_directory)
        source = source if source else self.lockdown.udid
        self._assert_backup_exists(backup_directory, source)

        with self.device_link(backup_directory) as dl, \
                NotificationProxyService(self.lockdown) as notification_proxy, \
                AfcService(self.lockdown) as afc:
            with self._backup_lock(afc, notification_proxy):
                manifest_plist_path = backup_directory / source / 'Manifest.plist'
                with open(manifest_plist_path, 'rb') as fd:
                    manifest = plistlib.load(fd)
                is_encrypted = manifest.get('IsEncrypted', False)
                options = {
                    'RestoreShouldReboot': reboot,
                    'RestoreDontCopyBackup': copy,
                    'RestorePreserveSettings': settings,
                    'RestoreSystemFiles': system,
                    'RemoveItemsNotRestored': remove,
                }
                if is_encrypted:
                    if password:
                        options['Password'] = password
                    else:
                        self.logger.error('Backup is encrypted, please supply password.')
                        return
                dl.send_process_message({
                    'MessageName': 'Restore',
                    'TargetIdentifier': self.lockdown.udid,
                    'SourceIdentifier': source,
                    'Options': options,
                })
                dl.dl_loop(progress_callback)

    def info(self, backup_directory='.', source: str = '') -> str:
        """
        Get information about a backup.
        :param backup_directory: Path of the backup directory.
        :param source: Identifier of device to get info about its backup.
        :return: Information about a backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.udid)
        with self.device_link(backup_dir) as dl:
            message = {'MessageName': 'Info', 'TargetIdentifier': self.lockdown.udid}
            if source:
                message['SourceIdentifier'] = source
            dl.send_process_message(message)
            result = dl.dl_loop()
        return result

    def list(self, backup_directory='.', source: str = '') -> str:
        """
        List the files in the last backup.
        :param backup_directory: Path of the backup directory.
        :param source: Identifier of device to list its backup data.
        :return: List of files and additional data about each file, all in a CSV format.
        """
        backup_dir = Path(backup_directory)
        source = source if source else self.lockdown.udid
        self._assert_backup_exists(backup_dir, source)
        with self.device_link(backup_dir) as dl:
            dl.send_process_message({
                'MessageName': 'List', 'TargetIdentifier': self.lockdown.udid, 'SourceIdentifier': source,
            })
            result = dl.dl_loop()
        return result

    def unback(self, backup_directory='.', password: str = '', source: str = '') -> None:
        """
        Unpack a complete backup to its device hierarchy.
        :param backup_directory: Path of the backup directory.
        :param password: Password of the backup if it is encrypted.
        :param source: Identifier of device to unpack its backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.udid)
        with self.device_link(backup_dir) as dl:
            message = {'MessageName': 'Unback', 'TargetIdentifier': self.lockdown.udid}
            if source:
                message['SourceIdentifier'] = source
            if password:
                message['Password'] = password
            dl.send_process_message(message)
            dl.dl_loop()

    def extract(self, domain_name: str, relative_path: str, backup_directory='.', password: str = '',
                source: str = '') -> None:
        """
        Extract a file from a previous backup.
        :param domain_name: File's domain name, e.g., SystemPreferencesDomain or HomeDomain.
        :param relative_path: File path.
        :param backup_directory: Path of the backup directory.
        :param password: Password of the last backup if it is encrypted.
        :param source: Identifier of device to extract file from its backup.
        """
        backup_dir = Path(backup_directory)
        self._assert_backup_exists(backup_dir, source if source else self.lockdown.udid)
        with self.device_link(backup_dir) as dl:
            message = {
                'MessageName': 'Extract', 'TargetIdentifier': self.lockdown.udid, 'DomainName': domain_name,
                'RelativePath': relative_path
            }
            if source:
                message['SourceIdentifier'] = source
            if password:
                message['Password'] = password
            dl.send_process_message(message)
            dl.dl_loop()

    def change_password(self, backup_directory='.', old: str = '', new: str = '') -> None:
        """
        Change backup password.
        :param backup_directory: Backups directory.
        :param old: Previous password. Omit when enabling backup encryption.
        :param new: New password. Omit when disabling backup encryption.
        """
        with self.device_link(Path(backup_directory)) as dl:
            message = {'MessageName': 'ChangePassword', 'TargetIdentifier': self.lockdown.udid}
            if old:
                message['OldPassword'] = old
            if new:
                message['NewPassword'] = new
            dl.send_process_message(message)
            dl.dl_loop()

    def erase_device(self, backup_directory='.') -> None:
        """
        Erase the device.
        """
        with suppress(ConnectionTerminatedError):
            with self.device_link(Path(backup_directory)) as dl:
                dl.send_process_message({'MessageName': 'EraseDevice', 'TargetIdentifier': self.lockdown.udid})
                dl.dl_loop()

    def version_exchange(self, dl: DeviceLink, local_versions=None) -> None:
        """
        Exchange versions with the device and assert that the device supports our version of the protocol.
        :param dl: Initialized device link.
        :param local_versions: versions supported by us.
        """
        if local_versions is None:
            local_versions = SUPPORTED_VERSIONS
        dl.send_process_message({
            'MessageName': 'Hello',
            'SupportedProtocolVersions': local_versions,
        })
        reply = dl.receive_message()
        assert reply[0] == 'DLMessageProcessMessage' and reply[1]['ErrorCode'] == 0
        assert reply[1]['ProtocolVersion'] in local_versions

    def init_mobile_backup_factory_info(self, afc: AfcService):
        with InstallationProxyService(self.lockdown) as ip, SpringBoardServicesService(self.lockdown) as sbs:
            root_node = self.lockdown.get_value()
            itunes_settings = self.lockdown.get_value(domain='com.apple.iTunes')
            min_itunes_version = self.lockdown.get_value('com.apple.mobile.iTunes', 'MinITunesVersion')
            app_dict = {}
            installed_apps = []
            apps = ip.browse(options={'ApplicationType': 'User'},
                             attributes=['CFBundleIdentifier', 'ApplicationSINF', 'iTunesMetadata'])
            for app in apps:
                bundle_id = app['CFBundleIdentifier']
                if bundle_id:
                    installed_apps.append(bundle_id)
                    if app.get('iTunesMetadata', False) and app.get('ApplicationSINF', False):
                        app_dict[bundle_id] = {
                            'ApplicationSINF': app['ApplicationSINF'],
                            'iTunesMetadata': app['iTunesMetadata'],
                            'PlaceholderIcon': sbs.get_icon_pngdata(bundle_id),
                        }

            files = {}
            for file in ITUNES_FILES:
                try:
                    data_buf = afc.get_file_contents('/iTunes_Control/iTunes/' + file)
                except AfcFileNotFoundError:
                    pass
                else:
                    files[file] = data_buf

            ret = {
                'iTunes Version': min_itunes_version if min_itunes_version else '10.0.1',
                'iTunes Files': files,
                'Unique Identifier': self.lockdown.udid.upper(),
                'Target Type': 'Device',
                'Target Identifier': root_node['UniqueDeviceID'],
                'Serial Number': root_node['SerialNumber'],
                'Product Version': root_node['ProductVersion'],
                'Product Type': root_node['ProductType'],
                'Installed Applications': installed_apps,
                'GUID': uuid.uuid4().bytes,
                'Display Name': root_node['DeviceName'],
                'Device Name': root_node['DeviceName'],
                'Build Version': root_node['BuildVersion'],
                'Applications': app_dict,
            }

            if 'IntegratedCircuitCardIdentity' in root_node:
                ret['ICCID'] = root_node['IntegratedCircuitCardIdentity']
            if 'InternationalMobileEquipmentIdentity' in root_node:
                ret['IMEI'] = root_node['InternationalMobileEquipmentIdentity']
            if 'MobileEquipmentIdentifier' in root_node:
                ret['MEID'] = root_node['MobileEquipmentIdentifier']
            if 'PhoneNumber' in root_node:
                ret['Phone Number'] = root_node['PhoneNumber']

            try:
                data_buf = afc.get_file_contents('/Books/iBooksData2.plist')
            except AfcFileNotFoundError:
                pass
            else:
                ret['iBooks Data 2'] = data_buf
            if itunes_settings:
                ret['iTunes Settings'] = itunes_settings
            return ret

    @contextmanager
    def _backup_lock(self, afc, notification_proxy):
        notification_proxy.notify_post(NP_SYNC_WILL_START)
        lockfile = afc.fopen('/com.apple.itunes.lock_sync', 'r+')
        if lockfile:
            notification_proxy.notify_post(NP_SYNC_LOCK_REQUEST)
            for _ in range(50):
                try:
                    afc.lock(lockfile, AFC_LOCK_EX)
                except AfcException as e:
                    if e.status == afc_error_t.OP_WOULD_BLOCK:
                        time.sleep(0.2)
                    else:
                        afc.fclose(lockfile)
                        raise
                else:
                    notification_proxy.notify_post(NP_SYNC_DID_START)
                    break
            else:  # No break, lock failed.
                afc.fclose(lockfile)
                raise PyMobileDevice3Exception('Failed to lock itunes sync file')
        try:
            yield
        finally:
            afc.lock(lockfile, AFC_LOCK_UN)
            afc.fclose(lockfile)
            notification_proxy.notify_post(NP_SYNC_DID_FINISH)

    @staticmethod
    def _assert_backup_exists(backup_directory: Path, identifier: str):
        device_directory = backup_directory / identifier
        assert (device_directory / 'Info.plist').exists()
        assert (device_directory / 'Manifest.plist').exists()
        assert (device_directory / 'Status.plist').exists()

    @contextmanager
    def device_link(self, backup_directory):
        dl = DeviceLink(self.service, backup_directory)
        dl.version_exchange()
        self.version_exchange(dl)
        try:
            yield dl
        finally:
            dl.disconnect()



import logging
import typing
from enum import Enum
from zipfile import ZipFile

from ipsw_parser.exceptions import NoSuchBuildIdentityError
from ipsw_parser.ipsw import IPSW

from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.restore.device import Device
from pymobiledevice3.restore.tss import TSSResponse

RESTORE_VARIANT_ERASE_INSTALL = 'Erase Install (IPSW)'
RESTORE_VARIANT_UPGRADE_INSTALL = 'Upgrade Install (IPSW)'
RESTORE_VARIANT_MACOS_RECOVERY_OS = 'macOS Customer'


class Behavior(Enum):
    Update = 'Update'
    Erase = 'Erase'


class BaseRestore:
    def __init__(self, ipsw: ZipFile, device: Device, tss: typing.Mapping = None,
                 behavior: Behavior = Behavior.Update, logger=None):
        self.logger = logging.getLogger(self.__class__.__name__) if logger is None else logger
        self.ipsw = IPSW(ipsw)
        self.device = device
        self.tss = TSSResponse(tss) if tss is not None else None

        if not self.device.is_image4_supported:
            raise NotImplementedError('is_image4_supported is False')

        self.logger.info(f'connected device: <ecid: {self.device.ecid} hardware_model: {self.device.hardware_model} '
                         f'image4-support: {self.device.is_image4_supported}>')

        self.logger.debug('scanning BuildManifest.plist for the correct BuildIdentity')

        variant = {
            Behavior.Update: RESTORE_VARIANT_UPGRADE_INSTALL,
            Behavior.Erase: RESTORE_VARIANT_ERASE_INSTALL,
        }[behavior]

        try:
            self.build_identity = self.ipsw.build_manifest.get_build_identity(self.device.hardware_model,
                                                                              restore_behavior=behavior.value,
                                                                              variant=variant)
        except NoSuchBuildIdentityError:
            if behavior == Behavior.Update:
                self.build_identity = self.ipsw.build_manifest.get_build_identity(self.device.hardware_model,
                                                                                  restore_behavior=behavior.value)
            else:
                raise

        self.macos_variant = None
        try:
            self.macos_variant = self.ipsw.build_manifest.get_build_identity(
                self.device.hardware_model,
                variant=RESTORE_VARIANT_MACOS_RECOVERY_OS)
            self.logger.info('Performing macOS restore')
        except NoSuchBuildIdentityError:
            pass

        build_info = self.build_identity.get('Info')
        if build_info is None:
            raise PyMobileDevice3Exception('build identity does not contain an "Info" element')

        device_class = build_info.get('DeviceClass')
        if device_class is None:
            raise PyMobileDevice3Exception('build identity does not contain an "DeviceClass" element')



import logging
import os
import platform
import plistlib
import sys
import uuid
from contextlib import suppress
from pathlib import Path
from typing import Mapping, Optional

from pymobiledevice3 import usbmux
from pymobiledevice3.common import get_home_folder
from pymobiledevice3.exceptions import MuxException, NotPairedError
from pymobiledevice3.usbmux import PlistMuxConnection

PAIR_RECORDS_PATH = {
    'win32': Path(os.environ.get('ALLUSERSPROFILE', ''), 'Apple', 'Lockdown'),
    'darwin': Path('/var/db/lockdown/'),
    'linux': Path('/var/lib/lockdown/'),
}

logger = logging.getLogger(__name__)


def generate_host_id(hostname: str = None) -> str:
    hostname = platform.node() if hostname is None else hostname
    host_id = uuid.uuid3(uuid.NAMESPACE_DNS, hostname)
    return str(host_id).upper()


def get_itunes_pairing_record(identifier: str) -> Optional[Mapping]:
    platform_type = 'linux' if not sys.platform.startswith('linux') else sys.platform
    filename = PAIR_RECORDS_PATH[platform_type] / f'{identifier}.plist'
    try:
        with open(filename, 'rb') as f:
            pair_record = plistlib.load(f)
    except (PermissionError, FileNotFoundError, plistlib.InvalidFileException):
        return None
    return pair_record


def get_local_pairing_record(identifier: str, pairing_records_cache_folder: Path) -> Optional[Mapping]:
    logger.debug('Looking for pymobiledevice3 pairing record')
    path = pairing_records_cache_folder / f'{identifier}.plist'
    if not path.exists():
        logger.debug(f'No pymobiledevice3 pairing record found for device {identifier}')
        return None
    return plistlib.loads(path.read_bytes())


def get_preferred_pair_record(identifier: str, pairing_records_cache_folder: Path) -> Mapping:
    """
    look for an existing pair record to connected device by following order:
    - usbmuxd
    - iTunes
    - local storage
    """

    # usbmuxd
    with suppress(NotPairedError, MuxException):
        with usbmux.create_mux() as mux:
            if isinstance(mux, PlistMuxConnection):
                pair_record = mux.get_pair_record(identifier)
    if pair_record is not None:
        return pair_record

    # iTunes
    pair_record = get_itunes_pairing_record(identifier)
    if pair_record is not None:
        return pair_record

    # local storage
    return get_local_pairing_record(identifier, pairing_records_cache_folder)


def create_pairing_records_cache_folder(pairing_records_cache_folder: Path = None) -> Path:
    if pairing_records_cache_folder is None:
        pairing_records_cache_folder = get_home_folder()
    else:
        pairing_records_cache_folder.mkdir(parents=True, exist_ok=True)
    return pairing_records_cache_folder
